(** Basic Utilities  *)

(** Chunxiao Lin @ 2006 *)

Require Import Bool.
Require Import List.

Set Implicit Arguments.
Unset Strict Implicit.

(** ******************************************************************* *)
(** tactics *)
(** ******************************************************************* *)
Ltac hypreplace H y :=
  let H' := fresh in
    (assert (H' := y); try (clear H; rename H' into H)).

Ltac elimbool :=
  match goal with 
    | [Handb : ?b1 && ?b2 = true |- _ ] =>
      let B1 := fresh "B" in let B2 := fresh "B" in
	(destruct (andb_prop _ _ Handb) as [B1 B2]; try (clear Handb); elimbool)
    | [Handb : true = ?b1 && ?b2 |- _ ] =>
      let B1 := fresh "B" in let B2 := fresh "B" in
	(destruct (andb_prop _ _ (sym_eq Handb)) as [B1 B2]; try (clear Handb); elimbool)

    | [Horb : ?b1 || ?b2 = true |- _ ] =>
      let B1 := fresh "B" in let B2 := fresh "B" in
	(destruct (orb_prop _ _ Horb) as [B1 | B2]; try (clear Horb); elimbool)
    | [Horb : true = ?b1 || ?b2 |- _ ] =>
      let B1 := fresh "B" in let B2 := fresh "B" in
	(destruct (orb_prop _ _ (sym_eq Horb)) as [B1 | B2]; try (clear Horb); elimbool)

    | [Horb : ?b1 || ?b2 = false |- _ ] =>
      let B1 := fresh "B" in let B2 := fresh "B" in
	(destruct (orb_false_elim _ _ Horb) as [B1  B2]; try (clear Horb); elimbool)
    | [Horb : true = ?b1 || ?b2 |- _ ] =>
      let B1 := fresh "B" in let B2 := fresh "B" in
	(destruct (orb_false_elim _ _ (sym_eq Horb)) as [B1 B2]; try (clear Horb); elimbool)

    | [Handb : true = ?b1 |- _ ] =>
      hypreplace Handb (sym_eq Handb); elimbool

    | _ => idtac
  end.

Ltac splitb := 
  match goal with 
    | [|- ?b1 && ?b2 = true] =>
      apply andb_true_intro; split; splitb 
    | [|- ?b1 || ?b2 = true] =>
      apply orb_true_intro
    | [|- ?b1 || ?b2 = false] =>
      apply orb_false_intro; splitb
    | _ => idtac
  end.

(** ******************************************************************* *)
(** bool *)
(** ******************************************************************* *)
Lemma negb_intro1 : forall b,
  b = false -> negb b = true.
Proof.
  intros; subst; reflexivity.
Qed.

Lemma negb_elim1 : forall b,
  negb b = true -> b = false.
Proof.
  intro b. 
  destruct b; simpl; auto.
Qed.

Lemma negb_intro2 : forall b,
  b = true -> negb b = false.
Proof.
  intros; subst; reflexivity.
Qed.

Lemma negb_elim2 : forall b,
  negb b = false -> b = true.
Proof.
  intro b. 
  destruct b; simpl; auto.
Qed.

Hint Resolve negb_intro1 negb_intro2 negb_elim1 negb_elim2 : natutil.

Lemma bool_dec : forall b, b = true \/ b = false.
Proof.
  intros; destruct b; firstorder.
Qed.


(** ******************************************************************* *)
(** list *)
(** ******************************************************************* *)
Fixpoint last (A : Set) (ll : list A) {struct ll} : option A :=
  match ll with
    | a :: nil => Some a
    | a :: ll' => last ll'
    | nil => None
  end.

Lemma last_next : forall (A : Set) (x y : A) l,
  last (x :: y :: l) = last (y :: l).
Proof.
  reflexivity.
Qed.

Lemma last_elim : forall (A : Set) ll (x: A),
  last ll = Some x -> In x ll.
Proof.
  induction ll; intros.
  discriminate.
  destruct ll.
  injection H; intros; subst.
  firstorder.
  rewrite last_next in H.
  firstorder.
Qed.

Lemma last_app : forall (A : Set) l1 l2 (x : A),
  last (l1 ++ (x :: l2)) = last (x :: l2).
Proof.
  induction l1; intros.
  reflexivity.
  rewrite <- app_comm_cons.
  destruct l1.
  reflexivity.
  rewrite <- app_comm_cons.
  rewrite last_next.
  hypreplace IHl1 (IHl1 l2 x).
  rewrite <- app_comm_cons in IHl1; trivial.
Qed.

Lemma list_last_dec : forall (A : Set) (ll : list A),
  ll = nil \/ exists l1, exists x, ll = l1 ++ x :: nil.
Proof.
  induction ll.
  left; reflexivity.
  right.
  destruct IHll.
  exists (nil (A := A)); exists a.
  subst; reflexivity.
  destruct H as [l1 [x Hl]].
  exists (a :: l1); exists x.
  rewrite <- app_comm_cons.
  subst; reflexivity.
Qed.

Lemma head_app_not_nil : forall (A: Set) (l1 l2 : list A),
  l1 <> nil -> head (l1 ++ l2) = head l1.
Proof.
  induction l1; intros.
  destruct H; reflexivity.
  reflexivity.
Qed.

Lemma app_x_nil_app : forall (A : Set) l1 l2 (x : A),
  (l1 ++ x :: nil) ++ l2 = l1 ++ (x :: l2).
Proof.
  induction l1; intros.
  reflexivity.
  repeat (rewrite <- app_comm_cons).
  rewrite (IHl1 l2 x); reflexivity.
Qed.


(** ******************************************************************* *)
(** nat list *)
(** ******************************************************************* *)
Lemma natlist_in_dec : forall (ll : list nat) x,
  In x ll \/ ~ In x ll.
Proof.
  induction ll; intros.
  firstorder.
  compare x a; intro e; subst.
  firstorder.
  destruct (IHll x);
    firstorder.
Qed.

Lemma natlist_break : forall ll (x : nat),
  In x ll -> 
  exists l1, exists l2, 
    ll = l1 ++ (x :: l2) /\ (~ In x l2).
Proof.
  induction ll; intros.
  destruct H.
  destruct (natlist_in_dec ll x).
  destruct (IHll _ H0) as [l1 [l2 [Happ Nin]]].
  exists (a :: l1); exists l2.
  split; trivial.
  rewrite <- app_comm_cons.
  inversion_clear Happ; reflexivity.
  simpl in H; destruct H.
  exists (nil (A := nat)); exists ll.
  subst; firstorder.
  tauto.
Qed.

Lemma natlist_break2 : forall ll (x : nat),
  In x ll -> 
  exists l1, exists l2, 
    ll = l1 ++ (x :: l2) /\ (~ In x l1).
Proof.
  induction ll; intros.
  destruct H.
  compare x a; intro e; subst.
  exists (nil (A := nat)); exists ll.
  firstorder.
  simpl in H; destruct H; subst.
  destruct e; reflexivity.
  destruct (IHll _ H) as [l1 [l2 [Happ Nin]]].
  exists (a :: l1); exists l2.
  split.
  rewrite <- app_comm_cons.
  inversion_clear Happ; reflexivity.
  intro F; apply Nin.
  destruct F; subst; trivial.
  destruct e; reflexivity.
Qed.


(** ******************************************************************* *)
(** Type list *)
(** ******************************************************************* *)
Inductive listT (A : Type) : Type :=
  | nilT : listT A
  | consT : A -> listT A -> listT A
.

Infix ":::" := consT (at level 60, right associativity).

Notation "( x ,, y ,, .. ,, z )" := (pairT .. (pairT x y) .. z).

Fixpoint InT (A : Type) (x : A) (ll : listT A) {struct ll} : Prop :=
  match ll with
    | nilT => False
    | consT y ll' => y = x \/ InT x ll'
  end.

Notation listT_sub := (fun ll ll' => (forall x, InT x ll -> InT x ll')).
