(** Basic machine model *)

(** Heap as ordered pairs *)

(** Chunxiao Lin @ 2006 *)

Require Export List.
Require Export Setoid.
Require Export Bool.

Require Export natutil.
Require Import basicutil.
Require Export natset2.

Set Implicit Arguments.
Unset Strict Implicit.

(* a well ordered heap implementation *)
(* a heap is a list of nat product plus a proof saying that 
  the list is well ordered and each address is multiple of 4 *)

(** ******************************************************************* *)
(** heap data *)
(** ******************************************************************* *)
Definition heap_data := (list (prod nat nat)).

(** ******************************************************************* *)
(** heap data in domain *)
(** ******************************************************************* *)
Fixpoint hd_in (n : nat) (s : heap_data) {struct s} : Prop :=
  match s with
    | nil => False
    | (x,y) :: s' => n = x \/ hd_in n s'
  end.

Fixpoint bhd_in (n : nat) (s : heap_data) {struct s} : bool :=
  match s with
    | nil => false
    | (x,y) :: s' => orb (beq_nat n x) (bhd_in n s')
  end.

Lemma hd_in_dec : forall s n, 
  hd_in n s \/ ~ hd_in n s.
Proof.
  induction s; simpl; intros.
  intuition.
  destruct a.
  destruct (IHs n); compare n n0; intros;
    intuition.
Qed.

Lemma bhd_in_dec : forall s n, 
  bhd_in n s = true \/ bhd_in n s = false.
Proof.
  induction s; simpl; intros.
  intuition.
  destruct a as [x y].
  destruct (IHs n); compare n x; intros.
  left; splitb; firstorder.
  left; splitb; firstorder.
  left; splitb; 
    left; subst n; apply beq_true.
  right; splitb; auto.
  apply beq_false; auto.
Qed.
  
Lemma bhd_in_elim1 : forall s n,
  bhd_in n s = true -> hd_in n s.
Proof.
  induction s; simpl; intros.
  discriminate H.
  destruct a as [x y].
  elimbool.
  left; rewrite (beq_nat_eq n x); auto.
  right; auto.
Qed.

Lemma bhd_in_elim2 : forall s n,
  bhd_in n s = false -> ~ hd_in n s.
Proof.
  induction s; simpl; intros.
  intro Hf; trivial.
  destruct a as [x y].
  elimbool.
  intro Hf; destruct Hf.
  subst n; rewrite (beq_true x) in B; discriminate B.
  apply (IHs _ B0 H).
Qed.

Lemma bhd_in_intro1 : forall s n,
    hd_in n s -> bhd_in n s = true.
Proof.
  induction s; simpl; intros.
  auto.
  destruct a as [x y].
  splitb.
  destruct H.
  left; subst n; rewrite (beq_true x); reflexivity.
  right; auto.
Qed.

Lemma bhd_in_intro2 : forall s n,
  ~ hd_in n s -> bhd_in n s = false.
Proof.
  induction s; simpl; intros.
  auto.
  destruct a as [x y].
  splitb; firstorder.
Qed.

Lemma hd_in_cons_false : forall s x y z,
    ~ hd_in x (cons (y,z) s) -> (x <> y /\ ~ hd_in x s).
Proof.
  intros.
  split; intro F; apply H;
    [left | right]; assumption.
Qed.


(** ******************************************************************* *)
(** order on heap data *)
(** ******************************************************************* *)

Definition lt_hd (n : nat) (s : heap_data) :=
  forall x, hd_in x s -> n < x.

Lemma lt_hd_weak : forall s a a', 
  a < a' -> lt_hd a' s -> lt_hd a s.
Proof.
  red; intros.
  hypreplace H0 (H0 x H1).
  omega.
Qed.

Lemma lt_hd_cons : forall s x a,
  lt_hd x (a :: s) -> lt_hd x s.
Proof.
  red; intros.
  apply H.
  destruct a as [n y].
  right; exact H0.
Qed.

Lemma lt_hd_not_in : forall s a,
  lt_hd a s -> ~ hd_in a s.
Proof.
  intros.
  intro I.
  hypreplace H (H _ I).
  omega.
Qed.

Fixpoint blt_hd (n : nat) (s : heap_data) {struct s} : bool :=
  match s with
    | nil => true
    | cons (x,y) s' => (blt_nat n x) && blt_hd n s'
  end.

Lemma blt_hd_elim : forall s n,
  blt_hd n s = true -> lt_hd n s.
Proof.
  induction s; simpl; intros n H x I.
  elimtype False; apply I.
  destruct a as [y z].
  elimbool.
  destruct I. 
  subst x.
  apply (blt_lt B).
  apply (IHs _ B0).
  exact H.
Qed.

Lemma blt_hd_intro : forall s n,
  lt_hd n s -> blt_hd n s = true.
Proof.
  induction s; simpl; intros.
  reflexivity.
  destruct a as [x y].
  splitb.
  apply blt_true.
  apply H.
  left; reflexivity.
  apply IHs.
  apply (lt_hd_cons H).
Qed.

Fixpoint bhd_canon (s : heap_data) : bool :=
  match s with
    | nil => true
    | (x,y) :: s' => bptr x && blt_hd x s' && bhd_canon s'
  end.


(** ******************************************************************* *)
(** heap *)
(** ******************************************************************* *)
Inductive heap : Set :=
  pfheap : forall s : heap_data, bhd_canon s = true -> heap.

Definition hd_of (s : heap) := let (hd, _) := s in hd.

Definition pf_of (s : heap) := 
  let (sd, pf) as s return (bhd_canon (hd_of s) = true) := s in pf.

(** ******************************************************************* *)
(** empty heap *)
(** ******************************************************************* *)
Definition emptypf : bhd_canon nil = true := (refl_equal true).

Definition empty_heap := pfheap emptypf.

(** ******************************************************************* *)
(** heap in domain *)
(** ******************************************************************* *)
Definition heap_in (n : nat) (s : heap) := hd_in n (hd_of s).

(** ******************************************************************* *)
(** heap get *)
(** ******************************************************************* *)
Fixpoint hd_get (s : heap_data) (n : nat) {struct s} : option nat :=
  match s with
    | nil => None
    | (x,y) :: s' => if (beq_nat x n) then Some y else hd_get s' n
  end.

Definition heap_get h addr :=
  hd_get (hd_of h) addr.

Lemma heap_get_ptr : forall h n v,
  heap_get h n = Some v -> ptr n.
Proof.
  intro h; destruct h; induction s as [| [x y] s'];
    unfold heap_get in * ; simpl in * ; intros.
  discriminate.
  generalize H; clear H.
  elimbool.
  compare x n; intro; subst; simplbeq; intro.
  injection H; intro; subst.
  apply bptr_elim; auto.
  apply IHs' with v; auto.
Qed.

Lemma heap_in_get_some : forall h n,
  heap_in n h -> exists y, heap_get h n = Some y.
Proof.
  destruct h; induction s as [|[x y] h']; 
    unfold heap_in in * ; unfold heap_get in * ; 
      simpl in * ;intros.
  elimtype False; apply H.
  destruct H; subst.
  exists y; simplbeq; reflexivity.
  destruct (andb_true_eq _ _ (sym_eq e)).
  destruct (IHh' (sym_eq H1) _ H).
  compare x n; intro; subst; simplbeq.
  exists y; reflexivity.
  exists x0; trivial.
Qed.

Lemma heap_get_dec : forall h a,
  heap_get h a = None \/ heap_get h a <> None.
Proof.
  intros.
  destruct (heap_get h a); [right | left].
  intro F; discriminate.
  reflexivity.
Qed.

Lemma heap_get_dec2 : forall h a,
  heap_get h a = None \/ exists v, heap_get h a = Some v.
Proof.
  intros.
  destruct (heap_get h a); [right; exists n | left];
    reflexivity.
Qed.

Lemma heap_get_elim1 : forall h v, heap_get h v = None -> ~ heap_in v h.
Proof.
  intros.
  intro F; destruct (heap_in_get_some F).
  rewrite H in H0; discriminate.
Qed.

Lemma heap_get_elim0 : forall h l v, heap_get h l = Some v -> heap_in l h.
Proof.
  destruct h.
  induction s as [| [ll vv] s].
  intros; discriminate.
  unfold heap_get; unfold heap_in in * ; simpl in * .
  intros l v; simplif; subst; intros.
  left; reflexivity.
  elimbool.
  right; apply IHs with B0 v; trivial.
Qed.

(** ******************************************************************* *)
(** heap set *)
(** ******************************************************************* *)
Fixpoint hd_insert' (n v : nat) (s : heap_data) {struct s} : heap_data :=
  match s with
    | nil => cons (n,v) s
    | cons (x,y) s' => if (blt_nat n x) then (n,v) :: s
      else if (beq_nat n x) then (n,v) :: s'
	else (x,y) :: (hd_insert' n v s')
  end.

Definition hd_insert (n v : nat) (s : heap_data) : heap_data :=
  if bptr n then hd_insert' n v s else s.

Lemma hd_insert'_intro1 : forall s a v,
  hd_in a (hd_insert' a v s).
Proof.
  induction s; simpl; intros.
  auto.
  destruct a as [x y].
  generalize (IHs a0); clear IHs.
  simplcmp a0 x.
  left; reflexivity.
  left. 
  reflexivity.
  simpl.
  right; auto.
Qed.

Lemma hd_insert'_intro2 : forall a b v s,
  hd_in a s ->
    hd_in a (hd_insert' b v s).
Proof.
  induction s; simpl; intros.
  tauto.
  destruct a0 as [x y].
  simplcmp b x; destruct H; simpl; auto.
  left.
  rewrite H; symmetry; apply beq_nat_eq; auto.
Qed.

Lemma hd_insert'_elim : forall a b v s,
  hd_in a (hd_insert' b v s) ->
  a = b \/ hd_in a s.
Proof.
  (*     unfold hd_insert. *)
  induction s as [ | c s];(*   destruct (bptr b); *) simpl;
    try tauto.
  destruct c as [x y].
  simplcmp b x; auto;
    intro H; destruct H; auto.
  destruct (IHs H); auto.
Qed.

Lemma lt_hd_insert' : forall s a a' v,
  a < a' -> lt_hd a s -> lt_hd a (hd_insert' a' v s).
Proof.
  red; intros.
  destruct (hd_insert'_elim H1).
  subst a'; exact H.
  apply (H0 _ H2).
Qed.

Lemma hd_insert_lt : forall s a v,
  ptr a -> lt_hd a s -> hd_insert a v s = (a,v) :: s.
Proof.
  unfold hd_insert.
  induction s as [ | c s];
    intros; rewrite bptr_intro; auto.
  hypreplace H0 (blt_hd_intro H0).
  destruct c as [x y].
  simpl in * .
  elimbool.
  rewrite B; reflexivity.
Qed.

Lemma hd_insert_canon : forall s a v,
    bhd_canon s = true ->
    bhd_canon (hd_insert a v s) = true.
Proof.
  unfold hd_insert.
  intros.
  
  destruct (bptr_dec a) as [R | R]; rewrite R.
  induction s as [_ | b s]; intros;
    simpl.
  auto with bool.
  
  destruct b as [x y].
  simpl in H.
  elimbool.
  hypreplace IHs (IHs B0).
  simpl.
  simplcmp a x.
  simpl.
  splitb; splitb; splitb; trivial.
  apply blt_hd_intro.
  apply lt_hd_weak with x.
  apply  (blt_lt RE).
  apply (blt_hd_elim B2).
  simpl.
  rewrite <- (beq_nat_eq  a x (sym_eq RE)) in B2.
  auto with bool.
  simpl.
  repeat (apply andb_true_intro; split; auto).
  apply blt_hd_intro.
  apply lt_hd_insert'.
  apply (blt_lt RE).
  apply (blt_hd_elim B2).
  assumption.
Qed.   
  

Lemma hd_set_get_eq' : forall h addr val,
  hd_get (hd_insert' addr val h) addr = Some val.
Proof.
  induction h; intros.
  simpl; simplbeq; reflexivity.
  destruct a as [x y].
  simpl.
  simplcmp addr x; simpl; simplbeq; auto.
  hypreplace RE (blt_lt RE).
  omega.
Qed.

Lemma hd_set_get_ne' : forall h addr addr' val,
  addr <> addr' -> hd_get (hd_insert' addr' val h) addr = hd_get h addr.
Proof.
  induction h; simpl; intros.
  simplbeq.
  reflexivity.
  destruct a as [x y].
  simplcmp addr' x; simpl.
  rewrite beq_false; auto.
  rewrite beq_false; auto.
  rewrite (beq_nat_eq addr' x (sym_eq RE)) in H.
  simplbeq; auto.
  destruct (beq_nat x addr); auto.
Qed.

Definition heap_set (h : heap) (addr : nat) (val : nat) := 
    pfheap (hd_insert_canon addr val (pf_of h)).

Lemma hd_set_get_eq : forall h addr val,
  ptr addr -> hd_get (hd_insert addr val h) addr = Some val.
Proof.
  induction h; intros;
    unfold hd_insert; rewrite bptr_intro; auto; simpl.
  simplbeq.
  reflexivity.
  destruct a as [x y].
  simplcmp addr x; simpl; simplbeq; auto.
  generalize (IHh _ val H).
  unfold hd_insert; rewrite bptr_intro; auto.
  hypreplace RE (blt_lt RE).
  omega.
Qed.

Lemma heap_set_get_eq : forall h addr val,
  ptr addr -> heap_get (heap_set h addr val) addr = Some val.
Proof.
  intros.
  destruct h.
  unfold heap_get; unfold heap_set.
  simpl.
  apply hd_set_get_eq; auto.
Qed.

Lemma hd_set_get_ne : forall h addr addr' val,
  addr <> addr' -> hd_get (hd_insert addr' val h) addr = hd_get h addr.
Proof.
  induction h; simpl; intros;
    try (hypreplace IHh (IHh _ _ val H));
    unfold hd_insert in * ; destruct (bptr addr'); auto; simpl.
  simplbeq.
  reflexivity.
  destruct a as [x y].
  simplcmp addr' x; simpl.
  rewrite beq_false; auto.
  rewrite beq_false; auto.
  rewrite (beq_nat_eq addr' x (sym_eq RE)) in H.
  simplbeq; auto.
  destruct (beq_nat x addr); auto.
Qed.

Lemma heap_set_get_ne : forall h addr addr' val,
  addr <> addr' -> heap_get (heap_set h addr' val) addr = heap_get h addr.
Proof.
  intros.
  destruct h.
  unfold heap_get; unfold heap_set.
  simpl.
  apply hd_set_get_ne; trivial.
Qed.


(** ******************************************************************* *)
(** singleton heap *)
(** ******************************************************************* *)
Definition singleton_heap addr val :=
  heap_set empty_heap addr val.


(** ******************************************************************* *)
(** heap remove *)
(** ******************************************************************* *)
Fixpoint hd_remove (n : nat) (h : heap_data) {struct h} : heap_data :=
  match h with
    | nil => h
    | (x, y) :: h' =>
      if (beq_nat n x) then hd_remove n h'
        else ((x, y) :: (hd_remove n h'))
  end. 

Lemma hd_remove_intro1 : forall s x,
  ~ hd_in x (hd_remove x s).
Proof.
  induction s; intros.
  tauto.
  simpl.
  destruct a as [a y].
  compare x a; intro e; try (subst x); simplbeq.
  exact (IHs a).
  intro H; destruct H.
  exact (e H).
  exact (IHs _ H).
Qed.

Lemma hd_remove_intro2 : forall s x y,
  x <> y -> hd_in x s -> hd_in x (hd_remove y s).
Proof.
  induction s; intros.
  tauto.
  simpl.
  destruct a as [a z].
  compare y a; intro e; try (subst y); simplbeq.
  destruct H0.
  elimtype False; exact (H H0).
  exact (IHs _ _ H H0).
  destruct H0; [left | right].
  exact H0.
  exact (IHs _ _ H H0).
Qed.

Lemma hd_remove_elim1 : forall s x y,
  hd_in x (hd_remove y s) -> x <> y. 
Proof.
  induction s; intros x y.
  tauto.
  simpl.
  destruct a as [a z].
  compare y a; intro e; subst; simplbeq; intro.
  firstorder.
  destruct H;  firstorder.
Qed.

Lemma hd_remove_elim2 : forall s x y,
  hd_in x (hd_remove y s) -> hd_in x s. 
Proof.
  induction s; intros x y.
  tauto.
  simpl.
  destruct a as [a z].
  compare y a; intro e; subst; simplbeq; intro.
  firstorder.
  destruct H;  firstorder.
Qed.

Lemma lt_hd_remove : forall s a a',
  lt_hd a s -> lt_hd a (hd_remove a' s).
Proof.
  red; intros.
  apply (H _  (hd_remove_elim2 H0)).
Qed.

Lemma hd_remove_none : forall s a,
  ~ hd_in a s -> hd_remove a s = s.
Proof.
  induction s as [ | b s]; intros.
  reflexivity.
  destruct b as [b x].
  destruct (hd_in_cons_false H).
  simpl; simplbeq.
  rewrite (IHs _ H1); reflexivity.
Qed.

Lemma hd_remove_canon : forall s a,
  bhd_canon s = true ->
  bhd_canon (hd_remove a s) = true.
Proof.
  induction s as [_ | b s]; intros.
  reflexivity.
  simpl in H.
  destruct b as [b x].
  elimbool.
  simpl.
  simplif; subst; auto.
  simpl; splitb; auto.
  splitb; auto.
  apply blt_hd_intro.
  apply lt_hd_remove.
  apply (blt_hd_elim B2).
Qed.

Definition heap_remove (h : heap) (n : nat) := 
  pfheap (hd_remove_canon n (pf_of h)).

Lemma heap_remove_intro1 : forall s x,
    ~ heap_in x (heap_remove s x).
Proof.
  destruct s; unfold heap_in; unfold heap_remove; 
    simpl; apply hd_remove_intro1.
Qed.

Lemma heap_remove_intro2 : forall s x y,
  x <> y -> heap_in x s ->
    heap_in x (heap_remove s y).
Proof.
  destruct s; unfold heap_in; unfold heap_remove; 
    simpl; apply hd_remove_intro2.
Qed.

Lemma heap_remove_elim1 : forall s x y,
  heap_in x (heap_remove s y) -> x <> y.
Proof.
  unfold heap_in; unfold heap_remove; 
    destruct s; simpl; apply hd_remove_elim1.
Qed.

Lemma heap_remove_elim2 : forall s x y,
  heap_in x (heap_remove s y) -> heap_in x s.
Proof.
  unfold heap_in; unfold heap_remove; 
    destruct s; simpl; apply hd_remove_elim2.
Qed.

Lemma hd_remove_get_eq : forall h addr,
  hd_get (hd_remove addr h) addr = None.
Proof.
  induction h; intros.
  reflexivity.
  destruct a as [a x].
  simpl.
  simplif; subst.
  apply IHh.
  simpl.
  simplbeq.
  apply IHh.
Qed.

Lemma heap_remove_get_eq : forall h addr,
  heap_get (heap_remove h addr) addr = None.
Proof.
  intros.
  destruct h.
  unfold heap_get; unfold heap_remove.
  simpl.
  apply hd_remove_get_eq; auto.
Qed.

Lemma hd_remove_get_ne : forall h addr addr',
  addr <> addr' -> hd_get (hd_remove addr' h) addr = hd_get h addr.
Proof.
  induction h; intros.
  reflexivity.
  destruct a as [a x].
  simpl.
  simplif; subst; auto.
  simplbeq; auto.
  simpl.
  simplif; auto.
Qed.

Lemma heap_remove_get_ne : forall h addr addr',
  addr <> addr' -> heap_get (heap_remove h addr') addr = heap_get h addr.
Proof.
  intros.
  destruct h.
  unfold heap_get; unfold heap_remove.
  simpl.
  apply hd_remove_get_ne; auto.
Qed.

(** ******************************************************************* *)
(** heap diff *)
(** ******************************************************************* *)
Fixpoint hd_diff (h1 h2 : heap_data) {struct h2} : heap_data :=
  match h2 with
    | nil => h1
    | (x, y) :: h2' => hd_diff (hd_remove x h1) h2'
  end. 

Lemma hd_diff_intro0 : forall h2 h1 x,
  ~ hd_in x h1 -> ~ hd_in x (hd_diff h1 h2).
Proof.
  induction h2; intros.
  tauto.
  simpl.
  destruct a as [a y].
  apply IHh2.
  intro F; apply H.
  apply (hd_remove_elim2 F).
Qed.

Lemma hd_diff_intro1 : forall h2 h1 x,
  hd_in x h2 -> ~ hd_in x (hd_diff h1 h2).
Proof.
  induction h2; intros.
  tauto.
  simpl.
  destruct a as [a y].
  compare x a; intro e; subst.
  apply hd_diff_intro0.
  apply hd_remove_intro1.
  simpl in H.
  destruct H; firstorder.
Qed.

Lemma hd_diff_intro2 : forall h2 h1 x,
  hd_in x h1 -> (~ hd_in x h2) -> hd_in x (hd_diff h1 h2).
Proof.
  induction h2; intros.
  tauto.
  simpl.
  destruct a as [a y].
  compare x a; intro e; subst; simpl in H0.
  firstorder.
  apply IHh2.
  apply hd_remove_intro2; trivial.
  firstorder.
Qed.

Lemma hd_diff_elim1 : forall h2 h1 x,
  hd_in x (hd_diff h1 h2) -> hd_in x h1. 
Proof.
  induction h2; intros.
  tauto.
  simpl.
  destruct a as [a y].
  apply (hd_remove_elim2 (IHh2 _ _ H)).
Qed.

Lemma hd_diff_elim2 : forall h2 h1 x,
  hd_in x (hd_diff h1 h2) -> ~ hd_in x h2. 
Proof.
  induction h2; intros.
  tauto.
  simpl.
  destruct a as [a y].
  intro F; destruct F; subst; simpl in H.
  apply (hd_remove_intro1 (hd_diff_elim1 H)).
  apply (IHh2 _ _ H H0).
Qed.

Lemma lt_hd_diff : forall h2 h1 a,
  lt_hd a h1 -> lt_hd a (hd_diff h1 h2).
Proof.
  red; intros.
  apply (H _ (hd_diff_elim1 H0)).
Qed.

Lemma hd_diff_nil : forall h,
  hd_diff nil h = nil.
Proof.
  induction h as [| [a b] h]; firstorder.
Qed.

Lemma hd_diff_canon : forall h2 h1,
  bhd_canon h1 = true ->
  bhd_canon (hd_diff h1 h2) = true.
Proof.
  induction h2 as [_ | [b x] h2]; intros.
  assumption.
  simpl.
  apply IHh2.
  apply hd_remove_canon; assumption.
Qed.

Definition heap_diff (h1 h2 : heap) := 
  pfheap (hd_diff_canon (hd_of h2) (pf_of h1)).

Lemma heap_diff_intro0 : forall h2 h1 x,
  ~ heap_in x h1 -> ~ heap_in x (heap_diff h1 h2).
Proof.
  destruct h1; unfold heap_in; unfold heap_diff; 
    simpl; apply hd_diff_intro0.
Qed.

Lemma heap_diff_intro1 : forall h2 h1 x,
  heap_in x h2 -> ~ heap_in x (heap_diff h1 h2).
Proof.
  destruct h1; unfold heap_in; unfold heap_diff; 
    simpl; apply hd_diff_intro1.
Qed.

Lemma heap_diff_intro2 : forall h2 h1 x,
  heap_in x h1 -> (~ heap_in x h2) -> heap_in x (heap_diff h1 h2).
Proof.
  destruct h1; unfold heap_in; unfold heap_diff; 
    simpl; apply hd_diff_intro2.
Qed.

Lemma heap_diff_elim1 : forall h2 h1 x,
  heap_in x (heap_diff h1 h2) -> heap_in x h1. 
Proof.
  destruct h1; unfold heap_in; unfold heap_diff; 
    simpl; apply hd_diff_elim1.
Qed.

Lemma heap_diff_elim2 : forall h2 h1 x,
  heap_in x (heap_diff h1 h2) -> ~ heap_in x h2. 
Proof.
  destruct h1; unfold heap_in; unfold heap_diff; 
    simpl; apply hd_diff_elim2.
Qed.

(** ******************************************************************* *)
(** heap append *)
(** ******************************************************************* *)
Fixpoint hd_append (s1 s2 : heap_data) {struct s1} : heap_data :=
  match s1 with
    | nil => s2
    | cons (x,y) s1' => hd_insert' x y (hd_append s1' s2)
  end.

Lemma hd_append_intro1 : forall a s1 s2,
  hd_in a s1 -> hd_in a (hd_append s1 s2).
Proof.
  induction s1; simpl.
  intuition.
  destruct a0 as [x y].
  destruct 1.
  subst x.
  apply hd_insert'_intro1.
  apply hd_insert'_intro2; auto.
Qed.

Lemma hd_append_intro2 : forall a s1 s2,
  hd_in a s2 -> hd_in a (hd_append s1 s2).
Proof.
  induction s1; simpl; intros.
  intuition.
  destruct a0 as [x y].
  apply hd_insert'_intro2; auto.
Qed.

Lemma hd_append_intro : forall a s1 s2,
  hd_in a s1 \/ hd_in a s2 -> hd_in a (hd_append s1 s2).
Proof.
  intros.
  destruct H.
  apply (hd_append_intro1 s2 H).
  apply (hd_append_intro2 s1 H).
Qed.

Lemma hd_append_elim : forall a s1 s2,
  hd_in a (hd_append s1 s2) -> hd_in a s1 \/ hd_in a s2.
Proof.
  induction s1; simpl; intros.
  tauto.
  destruct a0 as [x y].
  destruct (hd_insert'_elim H); auto.
  destruct (IHs1 _ H0); auto.
Qed.

Lemma hd_insert'_canon : forall s a v,
  ptr a -> bhd_canon s = true ->
  bhd_canon (hd_insert' a v s) = true.
Proof.
  induction s as [_ | b s]; intros;
    simpl.
  hypreplace H (bptr_intro H).
  auto with bool.
  
  destruct b as [x y].
  simpl in H0.
  elimbool.
  hypreplace IHs (IHs _ v H B0).
  simplcmp a x.
  simpl.
  repeat (apply andb_true_intro; split); trivial.
  apply (bptr_intro H).
  apply blt_hd_intro.
  apply lt_hd_weak with x.
  apply  (blt_lt RE).
  apply (blt_hd_elim B2).
  simpl.
  rewrite <- (beq_nat_eq  a x (sym_eq RE)) in B2.
  auto with bool.
  simpl.
  repeat (apply andb_true_intro; split; auto).
  apply (bptr_intro H).
  simpl.
  repeat (apply andb_true_intro; split; auto).
  apply blt_hd_intro.
  apply lt_hd_insert'.
  apply (blt_lt RE).
  apply (blt_hd_elim B2).
Qed.   

Lemma hd_append_canon : forall s1 s2,
  bhd_canon s1 = true ->
  bhd_canon s2 = true ->
  bhd_canon (hd_append s1 s2) = true.
Proof.
  induction s1 as [_ | b s]; intros.
  simpl; exact H0.
  simpl.
  destruct b as [x y].
  simpl in H.
  elimbool.
  apply hd_insert'_canon.
  apply bptr_elim; auto.
  apply IHs; assumption.
Qed.

Definition heap_append (s s' : heap) := 
  pfheap (hd_append_canon (pf_of s) (pf_of s')).


Lemma heap_append_intro : forall ha hb n v,
  heap_get ha n = Some v -> heap_get (heap_append ha hb) n = Some v.
Proof.
  destruct ha.
  induction s as [| [x y] s']; unfold heap_get in * ; 
    unfold heap_append in * ; simpl in * ; intros.
  discriminate.
  generalize H; clear H.
  compare x n; intro; subst; simplbeq; intro.
  rewrite <- H.
  elimbool.
  apply hd_set_get_eq'; trivial.
  elimbool.
  rewrite <- (IHs' B0 hb _ _ H).
  apply hd_set_get_ne'; auto. 
Qed.

Lemma heap_append_intro2 : forall ha hb n v,
  heap_get ha n = None -> heap_get hb n = Some v -> 
  heap_get (heap_append ha hb) n = Some v.
Proof.
  destruct ha.
  induction s as [| [x y] s']; unfold heap_get in * ; 
    unfold heap_append in * ; simpl in * ; intros.
  trivial.
  generalize H; clear H.
  compare x n; intro; subst; simplbeq; intro.
  discriminate.
  destruct (andb_true_eq _ _ (sym_eq e)).
  rewrite <- (IHs' (sym_eq H2) hb _ _ H H0).
  apply hd_set_get_ne'; auto. 
Qed.

Lemma heap_append_intro3 : forall ha hb n,
  heap_get ha n = None -> heap_get hb n = None -> 
  heap_get (heap_append ha hb) n = None.
Proof.
  destruct ha.
  induction s as [| [x y] s']; unfold heap_get in * ; 
    unfold heap_append in * ; simpl in * ; intros.
  trivial.
  generalize H; clear H.
  compare x n; intro; subst; simplbeq; intro.
  discriminate.
  destruct (andb_true_eq _ _ (sym_eq e)).
  rewrite <- (IHs' (sym_eq H2) _ _ H H0).
  apply hd_set_get_ne'; auto. 
Qed.


(** ******************************************************************* *)
(** heap eqv *)
(** ******************************************************************* *)
Definition heap_eqv (s1 s2 : heap) := 
  forall n, heap_get s1 n = heap_get s2 n.

Definition hd_eqv (s1 s2 : heap_data) := 
  forall n, hd_get s1 n = hd_get s2 n.

Lemma hd_eqv_elim : forall s1 s2,
  hd_eqv (hd_of s1) (hd_of s2) -> heap_eqv s1 s2.
Proof.
  tauto.
Qed.

Lemma hd_eqv_intro : forall s1 s2,
  heap_eqv s1 s2 -> hd_eqv (hd_of s1) (hd_of s2).
Proof.
  tauto.
Qed.

Lemma heap_eqv_refl : forall s, heap_eqv s s.
Proof.
  red; intros.
  tauto.
Qed.

Lemma heap_eqv_sym : forall s1 s2,
  heap_eqv s1 s2 -> heap_eqv s2 s1.
Proof.
  intros s1 s2 He n.
  hypreplace He (He n).
  auto.
Qed.

Lemma heap_eqv_trans : forall s1 s2 s3,
  heap_eqv s1 s2 -> heap_eqv s2 s3 -> heap_eqv s1 s3.
Proof.
  intros s1 s2 s3 e1 e2 n.
  hypreplace e1 (e1 n).
  hypreplace e2 (e2 n).
  rewrite e1; auto.
Qed.

(** ******************************************************************* *)
(** heap extensionality *)
(** ******************************************************************* *)
Require Import Classical_Prop.

Definition hd_eq (s s' : heap) := (hd_of s) = (hd_of s').

Lemma hd_eq_intro : forall s s',
  s = s' -> hd_eq s s'.
Proof.
  intros.
  destruct s; destruct s'. 
  injection H; intros; subst s0.
  rewrite H. 
  unfold hd_eq; reflexivity.
Qed.
  
Lemma hd_eq_elim : forall s s',
  hd_eq s s' -> s = s'.
Proof.
  intros.
  destruct s; destruct s'.
  unfold hd_eq in H.
  simpl in H.
  subst s0.
  assert (e = e0).
  apply proof_irrelevance.
  subst e0.
  reflexivity.
Qed.

Lemma hd_eqv_nil_eq : forall s,
  hd_eqv nil s -> s = nil.
Proof.
  destruct s; intros.
  reflexivity.
  destruct p as [x y].
  hypreplace H (H x).
  generalize H; simpl; simplbeq.
  intro; discriminate.
Qed.

Lemma heap_eqv_empty_eq : forall s,
  heap_eqv empty_heap s -> s = empty_heap.
Proof.
  intros.
  hypreplace H (hd_eqv_intro H).
  hypreplace H (hd_eqv_nil_eq H).
  apply hd_eq_elim.
  unfold empty_heap; unfold hd_eq; simpl.
  exact H.
Qed.
  
Lemma hd_eqv_sym : forall h1 h2,
  hd_eqv h1 h2 -> hd_eqv h2 h1.
Proof.
  red; intros; auto.
Qed.

Lemma blt_hd_get_none : forall h n,
  blt_hd n h = true -> hd_get h n = None.
Proof.
  intros.
  hypreplace H (blt_hd_elim H).
  induction h as[|[x y] h'].
  reflexivity.
  simpl.
  assert (n < x).
  apply H.
  left; reflexivity.
  simplbeq. 
  apply IHh'.
  red; intros.
  apply H.
  right; auto.
  omega.
Qed.

Lemma hd_get_some_hd_in : forall h n v,
  hd_get h n = Some v -> hd_in n h.
Proof.
  induction h as [| [x y] h']; intros.
  discriminate.
  generalize H; clear H; simpl.
  compare x n; intro; subst; simplbeq; intro; auto.
  right; apply (IHh' _ _ H).
Qed.

Theorem heap_ext : forall s s',
  heap_eqv s s' -> s = s'.
Proof.
  intro s.
  destruct s.
  induction s; intros.
  assert (pfheap (s:=nil) e = empty_heap).
  apply hd_eq_elim.
  unfold empty_heap; unfold hd_eq; reflexivity.
  rewrite H0 in H; rewrite H0.
  symmetry.
  apply (heap_eqv_empty_eq H).
  destruct s'.
  destruct a as [x y].
  apply hd_eq_elim.
  unfold hd_eq; simpl.
  assert (H1 := hd_eqv_intro H).
  simpl in H1.
  simpl in e.
  destruct (andb_true_eq _ _ (sym_eq e)).
  destruct s0 as [| [x' y'] s'].
  apply (hd_eqv_nil_eq (hd_eqv_sym H1)).
  generalize (H x); clear H.
  unfold heap_get; simpl; simplbeq.
  simpl in e0.
  destruct (andb_true_eq _ _ (sym_eq e0)).
  destruct (andb_true_eq _ _ H0).
  destruct (andb_true_eq _ _ H).
  compare x' x; intro; subst; simplbeq.
  intro e1; injection e1; intro; subst y'; clear e1.
  assert (hd_eqv s s').
  intro n.
  generalize (H1 n).
  simpl.
  compare x n; intro; subst; simplbeq; intros; auto.
  rewrite (blt_hd_get_none (sym_eq H5)).
  rewrite (blt_hd_get_none (sym_eq H7)).
  reflexivity.
  assert (heap_eqv (pfheap (sym_eq H2)) (pfheap (sym_eq H3))).
  apply hd_eqv_elim; simpl; trivial.
  hypreplace IHs (hd_eq_intro (IHs (sym_eq H2) _ H9)).
  unfold hd_eq in IHs; simpl in IHs; subst; reflexivity.
  intro F.
  hypreplace F (blt_hd_elim (sym_eq H7) (hd_get_some_hd_in (sym_eq F))).
  generalize (H1 x').
  simpl; repeat simplbeq; intro F2.
  hypreplace F2 (blt_hd_elim (sym_eq H5) (hd_get_some_hd_in F2)).
  elimtype False; omega.
Qed.

(** ******************************************************************* *)
(** heap size *)
(** ******************************************************************* *)
Definition heap_size h := length (hd_of h).

(** ******************************************************************* *)
(** heap disjoin *)
(** ******************************************************************* *)
Definition heap_disj (h1 h2 : heap) : Prop :=
  forall addr, heap_in addr h1 -> ~ heap_in addr h2.

(** ******************************************************************* *)
(** subheap *)
(** ******************************************************************* *)
Definition heap_subheap (h1 h2 : heap) : Prop :=
  forall addr v, heap_get h1 addr = Some v -> heap_get h2 addr = Some v.


(** ******************************************************************* *)
(** domain of the heap *)
(** ******************************************************************* *)
Fixpoint hd_dom (hd : heap_data) : set :=
  match hd with
    | nil => empty_set
    | (a, x) :: hd' => set_add a (hd_dom hd')
  end.

Notation heap_dom := (fun (h : heap) => hd_dom (hd_of h)).

(** ******************************************************************* *)
(** heap join *)
(** ******************************************************************* *)
Definition elem_join (l r a : option nat) : Prop :=
  match l, r, a with
    | None, None, None => True
    | Some v1, None, Some v2 => v1 = v2
    | None, Some v1, Some v2 => v1 = v2
    | _, _, _ => False
  end.

Definition heap_join (h1 h2 h : heap) : Prop :=
  forall addr, elem_join (heap_get h1 addr) 
    (heap_get h2 addr) (heap_get h addr).

Lemma elem_join_left : forall v, 
  elem_join v None v.
Proof.
  destruct v; unfold elem_join; trivial.
Qed.

Lemma elem_join_right : forall v,
  elem_join None v v.
Proof.
  destruct v; unfold elem_join; trivial.
Qed.
  
Lemma heap_join_sym : forall ha hb h,
  heap_join ha hb h -> heap_join hb ha h.
Proof.
  intros ha hb h J.
  red; intro.
  assert (H := J addr).
  destruct (heap_get ha addr);
    destruct (heap_get hb addr);
      destruct (heap_get h addr);
	  exact H.
Qed.

Lemma heap_join_elim1 : forall ha hb h, 
  heap_join ha hb h -> heap_disj ha hb.
Proof.
  intros ha hb h J.
  red.
  unfold heap_in.
  intros addr Ha Hb.
  assert (J' := J addr).
  destruct (heap_in_get_some Ha).
  destruct (heap_in_get_some Hb).
  destruct (heap_get ha addr);
    destruct (heap_get hb addr);
      destruct (heap_get h addr);
	auto; discriminate.
Qed.

Lemma heap_join_elim3 : forall ha hb h, 
  heap_eqv ha empty_heap ->
  heap_join ha hb h ->
  heap_eqv hb h.
Proof.
  intros ha hb h E J.
  red; intros.
  assert (J' := J n).
  rewrite (E n) in J'.
  destruct (heap_get hb n);
    destruct (heap_get h n);
      unfold heap_get in J';
	simpl in J';
	  subst; try reflexivity; tauto.
Qed.

Lemma heap_join_assoc : forall ha hb hc hab h,
  heap_join ha hb hab ->
  heap_join hab hc h ->
  exists hbc, heap_join hb hc hbc /\
    heap_join ha hbc h.
Proof.
  intros ha hb hc hab h Jab Jabc.
  exists (heap_append hb hc).
  split; intro addr;
    assert (F1 := heap_append_intro);
      hypreplace F1 (F1 hb hc addr);
      assert (F2 := heap_append_intro2);
	hypreplace F2 (F2 hb hc addr);
	assert (F3 := heap_append_intro3);
	  hypreplace F3 (F3 hb hc addr);
	  assert (Jab' := Jab addr);
	    assert (Jabc' := Jabc addr);
	      destruct (heap_get ha addr);
		destruct (heap_get hb addr);
		  destruct (heap_get hc addr);
		    destruct (heap_get hab addr);
		      destruct (heap_get h addr);
			trivial;
			  try congruence;
			    try (elimtype False; assumption);
			      try (rewrite F3; auto; constructor);
				try (rewrite (F1 n); auto; constructor);
				  try (rewrite (F2 n); auto; constructor).
  simpl in * .
  rewrite F3; auto; subst; reflexivity.
  simpl in * .
  rewrite (F1 n); auto; subst; reflexivity.
Qed.		

Lemma heap_join_sound_none : forall ha hb h addr,
  heap_join ha hb h ->
  heap_get ha addr = None -> 
  heap_get hb addr = None ->
  heap_get h addr = None.
Proof.
  intros ha hb h addr J Ha Hb.
  assert (J' := J addr).
  rewrite Ha in J'.
  rewrite Hb in J'.
  destruct (heap_get h addr).
  simpl in J'.
  elimtype False; exact J'.
  reflexivity.
Qed.
  
Lemma heap_join_sound_left : forall ha hb h addr v,
  heap_join ha hb h ->
  heap_get ha addr = Some v -> 
    heap_get h addr = Some v.
Proof.
  intros ha hb h addr v J Ha.
  assert (H := hd_get_some_hd_in Ha).
  assert (Hb' := (heap_join_elim1 J H)).
  unfold heap_in in Hb'.
  destruct (heap_get_dec hb addr).
  assert (J' := J addr).
  rewrite Ha in J'.
  rewrite H0 in J'.
  destruct (heap_get h addr);
    simpl in J'.
  subst v; reflexivity.
  elimtype False; exact J'.
  elimtype False; apply Hb'.
  unfold heap_get in H0.
  assert (exists n,hd_get (hd_of hb) addr = Some n).
  destruct (hd_get (hd_of hb) addr).
  exists n; reflexivity.
  tauto.
  destruct H1.
  eapply hd_get_some_hd_in.
  apply H1.
Qed.

Lemma heap_join_sound_right : forall ha hb h addr v,
  heap_join ha hb h ->
  heap_get hb addr = Some v ->
  heap_get h addr = Some v.
Proof.
  intros ha hb h addr v J Hb.
  assert (H := hd_get_some_hd_in Hb).
  assert (Ha := (heap_join_elim1 (heap_join_sym J) H)).
  unfold heap_in in Hb.
  destruct (heap_get_dec ha addr).
  assert (J' := J addr).
  rewrite H0 in J'.
  rewrite Hb in J'.
  destruct (heap_get h addr);
    simpl in J'.
  subst v; reflexivity.
  elimtype False; exact J'.
  elimtype False; apply Ha.
  unfold heap_get in H0.
  assert (exists n,hd_get (hd_of ha) addr = Some n).
  destruct (hd_get (hd_of ha) addr).
  exists n; reflexivity.
  tauto.
  destruct H1.
  unfold heap_in;  eapply hd_get_some_hd_in.
  apply H1.
Qed.

(** ******************************************************************* *)
(** tactics *)
(** ******************************************************************* *)
Ltac simplhp :=
  match goal with
    | [ |- context [(heap_get (heap_set ?h ?a ?v) ?a)] ] => 
      rewrite heap_set_get_eq; simplhp
    | [ H : ?a1 <> ?a2 |- context 
      [(heap_get (heap_set ?h ?a1 ?v) ?a2)] ] => 
      rewrite heap_set_get_ne; [simplhp | auto]
    | [ H : ?a2 <> ?a1 |- context 
      [(heap_get (heap_set ?h ?a1 ?v) ?a2)] ] => 
      rewrite heap_set_get_ne; [simplhp | auto]
    | _ => idtac
  end.

(** ******************************************************************* *)
(** other lemmas *)
(** ******************************************************************* *)
Lemma heap_remove_size_dcs : forall h x n,
  heap_in x h -> 
  heap_size h = S n ->
  heap_size (heap_remove h x) = n.
Proof.
  destruct h.
  induction s as [ | [a y] s]; unfold heap_size; intros.
  elimtype False; apply H.
  simpl in * .
  destruct (andb_prop _ _ e) as [B0 B1].
  destruct (andb_prop _ _ B0) as [B2 B3]; clear B0.
  hypreplace B3 (lt_hd_not_in (blt_hd_elim B3)).
  destruct H.
  subst; simplbeq.
  rewrite (hd_remove_none B3). 
  inversion_clear H0; reflexivity.
  simplif; subst.
  tauto.
  simpl.
  unfold heap_size in IHs.
  simpl in IHs.
  destruct n.
  destruct s.
  tauto.
  discriminate.
  rewrite (IHs B1 x n H).
  reflexivity.
  inversion_clear H0.
  reflexivity.
Qed.

Lemma heap_in_dec : forall h x,
  heap_in x h \/ ~ heap_in x h.
Proof.
  intros; unfold heap_in; apply hd_in_dec.
Qed.

Lemma heap_get_some_heap_in : forall h x v,
  heap_get h x = Some v -> heap_in x h.
Proof.
  destruct h; induction s as [|[x y] h']; 
    unfold heap_in in * ; unfold heap_get in * ; 
      simpl in * ;intros.
  discriminate.
  generalize H; clear H.
  simplif; subst; intros.
  left; reflexivity.
  right.
  destruct (andb_prop _ _ e).
  apply (IHh' H1 _ _ H).
Qed.


Lemma heap_size_0_empty_heap : forall h,
  heap_size h = 0 -> h = empty_heap.
Proof.
  unfold heap_size; unfold empty_heap.
  destruct h; induction s; simpl; intros;
    apply heap_ext; red; intro;   
      unfold heap_get; simpl.
  reflexivity.
  discriminate.
Qed.

Lemma heap_get_empty_heap_None : forall x,
  heap_get empty_heap x = None.
Proof.
  reflexivity.
Qed.

Lemma heap_size_0_get_None : forall h x,
  heap_size h = 0 -> heap_get h x = None.
Proof.
  intros.
  rewrite (heap_size_0_empty_heap H).
  reflexivity.
Qed.

Lemma heap_join_set_left : forall ha hb h l v v',
  heap_get ha l = Some v ->
  heap_join ha hb h -> 
  heap_join (heap_set ha l v') hb (heap_set h l v').
Proof.
  intros ha hb h l v v' Hl J addr.
  assert (A := J addr).
  assert (Hptr := heap_get_ptr Hl).
  compare addr l; intro e; subst.
  simplhp; trivial.
  rewrite Hl in A.
  destruct (heap_get hb l); firstorder.
  repeat (rewrite heap_set_get_ne; trivial).
Qed.

Lemma heap_join_get_left : forall ha hb h l v,
  heap_get ha l = Some v ->
  heap_join ha hb h -> 
  heap_get h l = Some v.
Proof.
  intros ha hb h l v Hl J.
  assert (A := J l).
  rewrite Hl in A.
  destruct (heap_get hb l); firstorder.
  destruct (heap_get h l); firstorder.
Qed.

Lemma heap_get_some_in : forall h l v,
  heap_get h l = Some v -> heap_in l h.
Proof.
  intro h; destruct h.
  induction s as [| [a x] s].
  intros; discriminate.
  unfold heap_get; unfold heap_in. 
  simpl; intros l v; simplif; intros.
  left; reflexivity.
  right; unfold heap_in in IHs; simpl in * .
  elimbool.
  apply (IHs B0 l v).
  apply H.
Qed.

Lemma heap_set_dom_eq : forall h l v v',
  heap_get h l = Some v ->
  heap_dom (heap_set h l v') = heap_dom h.
Proof.
  intro h; destruct h.
  induction s as [| [a x] s].
  intros; discriminate.
  unfold heap_get; unfold heap_set; unfold hd_insert.
  simpl.
  intros l v v'.
  simplif; intros.
  injection H; intros; subst.
  simpl in e; elimbool.
  rewrite B1.
  rewrite beq_not_blt; auto with natutil.
  simpl in e; elimbool.
  assert (A := heap_get_ptr (h := pfheap B0)).
  unfold heap_get in A; simpl in A.
  hypreplace A (A _ _ H).
  rewrite (bptr_intro A).
  simplcmp l a.
  hypreplace B2 (blt_hd_elim B2).
  assert (B := heap_get_some_in (h := pfheap B0)).
  unfold heap_get in B; simpl in B.
  hypreplace B (B _ _ H).
  hypreplace RE (blt_lt RE).
  hypreplace B (B2 _ B).
  elimtype False; omega.
  compare l a; intro e; subst.
  destruct R; reflexivity.
  rewrite (beq_false e) in RE.
  discriminate.
  simpl in IHs.
  rewrite <- (IHs B0 l v v' H).
  simpl.
  unfold hd_insert.
  rewrite (bptr_intro A).
  reflexivity.
Qed.

Ltac get2in := repeat
  (match goal with
     |[V : heap_get _ _ = None |-_] =>
       hypreplace V (heap_get_elim1 V)
     |[V : heap_get _ _ = Some _ |-_] =>
       hypreplace V (heap_get_elim0 V)
   end).

Lemma hd_remove_exchange : forall h l1 l2,
  hd_remove l1 (hd_remove l2 h) = hd_remove l2 (hd_remove l1 h).
Proof.
  induction h as [ | [a b] h]; intros.
  reflexivity.
  simpl.
  repeat simplif; subst.
  reflexivity.
  simpl; simplbeq.
  apply IHh.
  simpl; simplbeq.
  apply IHh.
  simpl.
  repeat simplif; subst.
  reflexivity.
  destruct R0; reflexivity.
  destruct R; reflexivity.
  rewrite IHh; reflexivity.
Qed.

Lemma hd_diff_remove_get : forall ha h l l' v,
  hd_get (hd_diff (hd_remove l h) ha) l' = Some v ->
  hd_get (hd_diff h ha) l' = Some v.
Proof.
  induction ha as [| [a b] ha]; simpl; intros.
  compare l l'; intro e; subst.
  rewrite hd_remove_get_eq in H; discriminate.
  rewrite hd_remove_get_ne in H; auto.
  apply IHha with l; trivial.
  rewrite hd_remove_exchange; trivial.
Qed.

Lemma heap_diff_subheap : forall ha h,
  heap_subheap (heap_diff h ha) h.
Proof.
  destruct ha. 
  induction s as [| [l v] s].
  firstorder.
  red; intros.
  unfold heap_get in * ; unfold heap_diff in * ; simpl in * .
  elimbool.
  apply (IHs B0 h addr v0).
  unfold heap_get; simpl.
  apply (hd_diff_remove_get H).
Qed.

Lemma heap_subheap_ex : forall ha h,
  heap_subheap ha h -> exists hb, heap_join ha hb h.
Proof.
  intros ha h Sub1.
  exists (heap_diff h ha).
  assert (Sub2 := heap_diff_subheap (ha := ha) (h :=h)).
  intro l.
  destruct (heap_get_dec2 ha l) as [Va | [va Va]];
  destruct (heap_get_dec2 (heap_diff h ha) l) as [Vb | [vb Vb]];
  destruct (heap_get_dec2 h l) as [V | [v V]]; 
    rewrite Va; rewrite Vb; rewrite V; simpl; auto.
  get2in; apply Vb; apply heap_diff_intro2; trivial.
  rewrite (Sub2 _ _ Vb) in V; discriminate.
  rewrite (Sub2 _ _ Vb) in V; inversion_clear V; reflexivity.
  rewrite (Sub1 _ _ Va) in V; discriminate.
  rewrite (Sub1 _ _ Va) in V; inversion_clear V; reflexivity.
  rewrite (Sub1 _ _ Va) in V; inversion_clear V; reflexivity.
  get2in.
  destruct (heap_diff_elim2 Vb); tauto.
Qed.