(** Ordered finite set *)

(** Chunxiao Lin @ 2006 *)

Require Export List.
Require Export Setoid.
Require Export Bool.

Require Export natutil.
Require Import basicutil.

Set Implicit Arguments.
Unset Strict Implicit.

(** a well ordered nat set implementation *)
(** a set is a list of nat plus a proof saying that the list is well ordered *)

(** ******************************************************************* *)
(** set data *)
(** ******************************************************************* *)
Definition set_data := (list nat).

(** ******************************************************************* *)
(** set data in domain *)
(** ******************************************************************* *)
Section SetDataIn.

  Fixpoint sd_in (n : nat) (s : set_data) {struct s} : Prop :=
    match s with
      | nil => False
      | x :: s' => n = x \/ sd_in n s'
    end.

  Fixpoint bsd_in (n : nat) (s : set_data) {struct s} : bool :=
    match s with
      | nil => false
      | x :: s' => orb (beq_nat n x) (bsd_in n s')
    end.

  Lemma sd_in_dec : forall s n, 
    sd_in n s \/ ~ sd_in n s.
  Proof.
    induction s; simpl; intros.
    intuition.
    destruct (IHs n); compare n a; intros;
      intuition.
  Qed.

  Lemma bsd_in_dec : forall s n, 
    bsd_in n s = true \/ bsd_in n s = false.
  Proof.
    induction s; simpl; intros.
    intuition.
    destruct (IHs n); compare n a; intros.
    left; apply orb_true_intro.
    right; exact H.
    left; apply orb_true_intro.
    right; exact H.
    left; apply orb_true_intro.
    left; subst n; apply beq_true.
    right; apply orb_false_intro; auto.
    apply beq_false; auto.
  Qed.

  Lemma bsd_in_elim1 : forall s n,
    bsd_in n s = true -> sd_in n s.
  Proof.
    induction s; simpl; intros.
    discriminate H.
    destruct (orb_prop _ _ H).
    left; rewrite (beq_nat_eq n a); auto.
    right; auto.
  Qed.

  Lemma bsd_in_elim2 : forall s n,
    bsd_in n s = false -> ~ sd_in n s.
  Proof.
    induction s; simpl; intros.
    intro Hf; trivial.
    destruct (orb_false_elim _ _ H).
    intro Hf; destruct Hf.
    subst n; rewrite (beq_true a) in H0; discriminate H0.
    apply (IHs _ H1 H2).
  Qed.

  Lemma bsd_in_intro1 : forall s n,
    sd_in n s -> bsd_in n s = true.
  Proof.
    induction s; simpl; intros.
    auto.
    apply orb_true_intro.
    destruct H.
    left; subst n; rewrite (beq_true a); reflexivity.
    right; auto.
  Qed.

  Lemma bsd_in_intro2 : forall s n,
    ~ sd_in n s -> bsd_in n s = false.
  Proof.
    induction s; simpl; intros.
    auto.
    apply orb_false_intro; intuition.
  Qed.

  Lemma sd_in_cons_false : forall s x y,
    ~ sd_in x (cons y s) -> (x <> y /\ ~ sd_in x s).
  Proof.
    intros.
    split; intro F; apply H;
      [left | right]; assumption.
  Qed.

End SetDataIn.


(** ******************************************************************* *)
(** order on set data *)
(** ******************************************************************* *)

Section Order.

  Definition lt_sd (n : nat) (s : set_data) :=
    forall x, sd_in x s -> n < x.

  Lemma lt_sd_weak : forall s a a', 
    a < a' -> lt_sd a' s -> lt_sd a s.
  Proof.
    red; intros.
    hypreplace H0 (H0 x H1).
    omega.
  Qed.

  Lemma lt_sd_cons : forall s a a',
    lt_sd a (a' :: s) -> lt_sd a s.
  Proof.
    red; intros.
    apply H.
    right; exact H0.
  Qed.

  Lemma lt_sd_not_in : forall s a,
    lt_sd a s -> ~ sd_in a s.
  Proof.
    intros.
    intro I.
    hypreplace H (H _ I).
    omega.
  Qed.

  Fixpoint blt_sd (n : nat) (s : set_data) {struct s} : bool :=
    match s with
      | nil => true
      | cons x s' => (blt_nat n x) && blt_sd n s'
    end.

  Lemma blt_sd_elim : forall s n,
    blt_sd n s = true -> lt_sd n s.
  Proof.
    induction s; simpl; intros n H x I.
    elimtype False; apply I.
    destruct I;
      destruct (andb_true_eq _ _ (sym_eq H)).
    subst x.
    apply (blt_lt (sym_eq H1)).
    apply (IHs _ (sym_eq H2)).
    exact H0.
  Qed.

  Lemma blt_sd_intro : forall s n,
    lt_sd n s -> blt_sd n s = true.
  Proof.
    induction s; simpl; intros.
    reflexivity.
    apply andb_true_intro.
    split.
    apply blt_true.
    apply H.
    left; reflexivity.
    apply IHs.
    apply (lt_sd_cons H).
  Qed.

  Fixpoint bsd_canon (s : set_data) : bool :=
    match s with
      | nil => true
      | n :: s' => blt_sd n s' && bsd_canon s'
    end.

End Order.


(** ******************************************************************* *)
(** set *)
(** ******************************************************************* *)
Section ProofSet.

  Inductive set : Set :=
    pfset : forall s : set_data, bsd_canon s = true -> set.
  
  Definition sd_of (s : set) := let (sd, _) := s in sd.

  Definition pf_of (s : set) := 
    let (sd, pf) as s return (bsd_canon (sd_of s) = true) := s in pf.

  Definition emptypf : bsd_canon nil = true := (refl_equal true).
  
  Definition empty_set := pfset emptypf.

  Definition set_in (n : nat) (s : set) := sd_in n (sd_of s).

End ProofSet.


(** ******************************************************************* *)
(** set add *)
(** ******************************************************************* *)
Section SetAdd.

  Fixpoint sd_insert (n : nat) (s : set_data) {struct s} : set_data :=
    match s with
      | nil => cons n s
      | cons x s' => if (blt_nat n x) then n :: s
	else if (beq_nat n x) then s
	else x :: (sd_insert n s')
    end.

  Lemma sd_insert_intro1 : forall a s,
    sd_in a (sd_insert a s).
  Proof.
    induction s; simpl.
    auto.
    simplcmp a a0.
    left; reflexivity.
    left; apply (beq_nat_eq _ _ (sym_eq RE)).
    right; assumption.
  Qed.
  
  Lemma sd_insert_intro2 : forall a b s,
    sd_in a s ->
    sd_in a (sd_insert b s).
  Proof.
    induction s; simpl; intros.
    auto.
    simplcmp b a0; destruct H; simpl; auto.
  Qed.

  Lemma sd_insert_elim : forall a b s,
    sd_in a (sd_insert b s) ->
    a = b \/ sd_in a s.
  Proof.
    induction s as [ | c s]; simpl.
    tauto.
    simplcmp b c; auto.
    intro H; destruct H; auto.
    destruct (IHs H); auto.
  Qed.

  Lemma lt_sd_insert : forall s a a',
    a < a' -> lt_sd a s -> lt_sd a (sd_insert a' s).
  Proof.
    red; intros.
    destruct (sd_insert_elim H1).
    subst a'; exact H.
    apply (H0 _ H2).
  Qed.

  Lemma sd_insert_lt : forall s a,
    lt_sd a s -> sd_insert a s = a :: s.
  Proof.
    induction s as [ | b s]; intros.
    reflexivity.
    hypreplace H (blt_sd_intro H).
    simpl in * .
    destruct (andb_true_eq _ _ (sym_eq H)).
    rewrite <- H0; reflexivity.
  Qed.

  Lemma sd_insert_canon : forall s a,
    bsd_canon s = true ->
    bsd_canon (sd_insert a s) = true.
  Proof.
    induction s as [_ | b s]; intros.
    reflexivity.
    simpl in H.
    destruct (andb_true_eq _ _ (sym_eq H)).
    hypreplace IHs (IHs a (sym_eq H1)).
    simpl.
    simplcmp a b.
    simpl.
    repeat (apply andb_true_intro; split); trivial.
    apply blt_sd_intro.
    apply lt_sd_weak with b.
    apply  (blt_lt RE).
    apply (blt_sd_elim (sym_eq H0)).
    symmetry; exact H0.
    symmetry; exact H1.
    apply H.
    simpl.
    apply andb_true_intro; split; trivial.
    apply blt_sd_intro.
    apply lt_sd_insert.
    apply (blt_lt RE).
    apply (blt_sd_elim (sym_eq H0)).
  Qed.   
  
  Definition set_add (n : nat) (s : set) := 
    pfset (sd_insert_canon n (pf_of s)).

  Lemma set_add_intro1 : forall a s,
    set_in a (set_add a s).
  Proof.
    destruct s; unfold set_in; unfold set_add; 
      simpl; apply sd_insert_intro1.
  Qed.
  
  Lemma set_add_intro2 : forall a b s,
    set_in a s ->
    set_in a (set_add b s).
  Proof.
    destruct s; unfold set_in; unfold set_add; 
      simpl; apply sd_insert_intro2.
  Qed.

  Lemma set_add_elim : forall a b s,
    set_in a (set_add b s) ->
    a = b \/ set_in a s.
  Proof.
    unfold set_in; unfold set_add; 
      destruct s; simpl; apply sd_insert_elim.
  Qed.

End SetAdd.


(** ******************************************************************* *)
(** set remove *)
(** ******************************************************************* *)
Section SetRemove.

  Fixpoint sd_remove (n : nat) (s : set_data) {struct s} : set_data :=
    match s with
      | nil => s
      | x :: s' => 
	if (beq_nat n x) then sd_remove n s'
	  else (x :: (sd_remove n s'))
    end.

  Lemma sd_remove_intro1 : forall s x,
    ~ sd_in x (sd_remove x s).
  Proof.
    induction s; intros.
    tauto.
    simpl.
    compare x a; intro e; try (subst x); simplbeq.
    exact (IHs a).
    intro H; destruct H.
    exact (e H).
    exact (IHs _ H).
  Qed.

  Lemma sd_remove_intro2 : forall s x y,
    x <> y -> sd_in x s -> sd_in x (sd_remove y s).
  Proof.
    induction s; intros.
    tauto.
    simpl.
    compare y a; intro e; try (subst y); simplbeq.
    destruct H0.
    elimtype False; exact (H H0).
    exact (IHs _ _ H H0).
    destruct H0; [left | right].
    exact H0.
    exact (IHs _ _ H H0).
  Qed.

  Lemma sd_remove_elim : forall s x y,
    sd_in x (sd_remove y s) -> (x <> y /\ sd_in x s). 
  Proof.
    induction s; intros x y.
    tauto.
    simpl.
    compare y a; intro e; try (subst y); simplbeq; intro.
    destruct (IHs _ _ H).
    split; [idtac | right]; assumption.
    destruct H.
    subst x; split; [idtac | left]; auto.
    destruct (IHs _ _ H).
    split; [idtac | right]; assumption.
  Qed.

  Lemma lt_sd_remove : forall s a a',
    lt_sd a s -> lt_sd a (sd_remove a' s).
  Proof.
    red; intros.
    destruct (sd_remove_elim H0).
    apply (H _  H2).
  Qed.

  Lemma sd_remove_none : forall s a,
    ~ sd_in a s -> sd_remove a s = s.
  Proof.
    induction s as [ | b s]; intros.
    reflexivity.
    destruct (sd_in_cons_false H).
    simpl; simplbeq.
    rewrite (IHs _ H1); reflexivity.
  Qed.

  Lemma sd_insert_remove : forall s a,
    sd_in a s -> bsd_canon s = true ->
    sd_insert a (sd_remove a s) = s.
  Proof.
    induction s as [ | b s]; intros.
    elimtype False; apply H.
    simpl.
    simpl in H0.
    destruct (andb_true_eq _ _ (sym_eq H0)).
    destruct H.
    subst b; simplbeq.
    rewrite (sd_remove_none (lt_sd_not_in (blt_sd_elim (sym_eq H1)))).
    rewrite (sd_insert_lt (blt_sd_elim (sym_eq H1))).
    reflexivity.
    fold sd_in in H.
    compare a b; intro e; [subst b | idtac]; simplbeq.
    elimtype False.
    apply (lt_sd_not_in (blt_sd_elim (sym_eq H1)) H).
    simpl.
    hypreplace H1 ((blt_sd_elim (sym_eq H1)) _ H).
    simplcmp a b.
    hypreplace RE (blt_lt RE).
    elimtype False; omega.
    hypreplace RE (beq_nat_eq _ _ (sym_eq RE)).
    elimtype False; omega.
    rewrite (IHs _ H (sym_eq H2)).
    reflexivity.
  Qed.

  Lemma sd_remove_canon : forall s a,
    bsd_canon s = true ->
    bsd_canon (sd_remove a s) = true.
  Proof.
    induction s as [_ | b s]; intros.
    reflexivity.
    simpl in H.
    destruct (andb_true_eq _ _ (sym_eq H)).
    hypreplace IHs (IHs a (sym_eq H1)).
    hypreplace H0 (sym_eq H0).
    hypreplace H1 (sym_eq H1).
    simpl.
    destruct (beq_dec a b); rewrite H2; trivial.
    simpl.
    apply andb_true_intro; split; trivial.
    apply blt_sd_intro.
    apply lt_sd_remove.
    apply (blt_sd_elim H0).
  Qed.

  Definition set_remove (n : nat) (s : set) := 
    pfset (sd_remove_canon n (pf_of s)).

  Lemma set_remove_intro1 : forall s x,
    ~ set_in x (set_remove x s).
  Proof.
    destruct s; unfold set_in; unfold set_remove; 
      simpl; apply sd_remove_intro1.
  Qed.
  
  Lemma set_remove_intro2 : forall s x y,
    x <> y -> set_in x s ->
    set_in x (set_remove y s).
  Proof.
    destruct s; unfold set_in; unfold set_remove; 
      simpl; apply sd_remove_intro2.
  Qed.

  Lemma set_remove_elim : forall s x y,
    set_in x (set_remove y s) ->
    (x <> y /\ set_in x s).
  Proof.
    unfold set_in; unfold set_remove; 
      destruct s; simpl; apply sd_remove_elim.
  Qed.

End SetRemove.


(** ******************************************************************* *)
(** set union *)
(** ******************************************************************* *)
Section SetUnion.

  Fixpoint sd_append (s1 s2 : set_data) {struct s1} : set_data :=
    match s1 with
      | nil => s2
      | cons x s1' => sd_insert x (sd_append s1' s2)
    end.

  Lemma sd_append_intro1 : forall a s1 s2,
    sd_in a s1 -> sd_in a (sd_append s1 s2).
  Proof.
    induction s1; simpl.
    intuition.
    destruct 1.
    subst a0.
    apply sd_insert_intro1.
    apply sd_insert_intro2; auto.
  Qed.

  Lemma sd_append_intro2 : forall a s1 s2,
    sd_in a s2 -> sd_in a (sd_append s1 s2).
  Proof.
    induction s1; simpl; intros.
    intuition.
    apply sd_insert_intro2; auto.
  Qed.

  Lemma sd_append_intro : forall a s1 s2,
    sd_in a s1 \/ sd_in a s2 -> sd_in a (sd_append s1 s2).
  Proof.
    intros.
    destruct H.
    apply (sd_append_intro1 s2 H).
    apply (sd_append_intro2 s1 H).
  Qed.

  Lemma sd_append_elim : forall a s1 s2,
    sd_in a (sd_append s1 s2) -> sd_in a s1 \/ sd_in a s2.
  Proof.
    induction s1; simpl; intros.
    tauto.
    destruct (sd_insert_elim H); auto.
    destruct (IHs1 _ H0); auto.
  Qed.

  Lemma sd_append_canon : forall s1 s2,
    bsd_canon s1 = true ->
    bsd_canon s2 = true ->
    bsd_canon (sd_append s1 s2) = true.
  Proof.
    induction s1 as [_ | b s]; intros.
    simpl; exact H0.
    simpl.
    apply sd_insert_canon.
    destruct (andb_true_eq _ _ (sym_eq H)).
    hypreplace H2 (sym_eq H2).
    apply IHs; assumption.
  Qed.

  Definition set_union (s s' : set) := 
    pfset (sd_append_canon (pf_of s) (pf_of s')).

  Lemma set_union_intro1 : forall a s1 s2,
    set_in a s1 -> set_in a (set_union s1 s2).
  Proof.
    unfold set_union; unfold set_in; 
      destruct s1; destruct s2; simpl;
	apply sd_append_intro1.
  Qed.

  Lemma set_union_intro2 : forall a s1 s2,
    set_in a s2 -> set_in a (set_union s1 s2).
  Proof.
    unfold set_union; unfold set_in; 
      destruct s1; destruct s2; simpl;
	apply sd_append_intro2.
  Qed.

  Lemma set_union_intro : forall a s1 s2,
    set_in a s1 \/ set_in a s2 -> set_in a (set_union s1 s2).
  Proof.
    unfold set_union; unfold set_in; 
      destruct s1; destruct s2; simpl;
	apply sd_append_intro.
  Qed.

  Lemma set_union_elim : forall a s1 s2,
    set_in a (set_union s1 s2) -> set_in a s1 \/ set_in a s2.
  Proof.
    unfold set_union; unfold set_in; 
      destruct s1; destruct s2; simpl;
	apply sd_append_elim.
  Qed.

End SetUnion.


(** ******************************************************************* *)
(** set diff *)
(** ******************************************************************* *)
Section SetDiff.

  Fixpoint sd_diff (s1 s2 : set_data) {struct s2} : set_data :=
    match s2 with
      | nil => s1
      | cons x s2' => sd_remove x (sd_diff s1 s2')
    end.

  Lemma sd_diff_intro : forall a s1 s2,
    sd_in a s1 -> ~sd_in a s2 ->
    sd_in a (sd_diff s1 s2).
  Proof.
    induction s2; intros.
    exact H.
    destruct (sd_in_cons_false H0).
    simpl.
    apply sd_remove_intro2; trivial.
    apply IHs2; assumption.
  Qed.

  Lemma sd_diff_elim : forall a s1 s2,
    sd_in a (sd_diff s1 s2) ->
    sd_in a s1 /\ ~sd_in a s2.
  Proof.
    induction s2; simpl; intros.
    intuition.
    hypreplace H (sd_remove_elim H).
    intuition.
  Qed.

  Lemma sd_diff_canon : forall s1 s2,
    bsd_canon s1 = true ->
    bsd_canon (sd_diff s1 s2) = true.
  Proof.
    induction s2 as [_ | b s]; intros.
    simpl; exact H.
    simpl.
    apply sd_remove_canon; auto.
  Qed.

  Definition set_diff (s s' : set) := 
    pfset (sd_diff_canon (sd_of s') (pf_of s)).

  Lemma set_diff_intro : forall a s1 s2,
    set_in a s1 -> ~set_in a s2 ->
    set_in a (set_diff s1 s2).
  Proof.
    unfold set_diff; unfold set_in;
      destruct s1; destruct s2; simpl;
	apply sd_diff_intro.
  Qed.

  Lemma set_diff_elim : forall a s1 s2,
    set_in a (set_diff s1 s2) ->
    set_in a s1 /\ ~set_in a s2.
  Proof.
    unfold set_diff; unfold set_in;
      destruct s1; destruct s2; simpl;
	apply sd_diff_elim.
  Qed.

End SetDiff.


(** ******************************************************************* *)
(** set map *)
(** ******************************************************************* *)
Section SetMap.

  Lemma sd_map_canon : forall (f : nat->nat) s,
    bsd_canon s = true ->
    (forall x y, x < y -> f x < f y) ->
    bsd_canon (map f s) = true.
  Proof.
    induction s as [_ | b s]; intros.
    simpl; exact H.
    simpl.
    simpl in H.
    destruct (andb_true_eq _ _ (sym_eq H)); clear H.
    apply andb_true_intro; split.
    clear IHs H2.
    induction s.
    reflexivity.
    simpl in * .
    destruct (andb_true_eq _ _ H1); clear H1.
    apply andb_true_intro; split.
    apply (blt_true (H0 b a (blt_lt (sym_eq H)))).
    apply (IHs H2).
    apply IHs; auto.
  Qed.

  Definition set_map (f: nat -> nat) (s : set) 
    (PF : forall x y, x < y -> f x < f y) := 
    pfset (sd_map_canon (pf_of s) PF).

  Lemma set_map_elim : forall s a f (PF : forall x y, x < y -> f x < f y),
    set_in a (set_map s PF) ->
    exists b, a = f b /\ set_in b s.
  Proof.
    destruct s. 
    induction s as [ | x s]; simpl; intros.
    inversion_clear H.
    unfold set_map in H; unfold set_in in H.
    simpl in H.
    destruct H as [H | H].
    exists x; split; trivial.
    unfold set_in; simpl; left; reflexivity.
    destruct (andb_true_eq _ _ (sym_eq e)).
    fold bsd_canon in H1.
    destruct (IHs (sym_eq H1) a _ PF) as [b [B1 B2]].
    apply H.
    exists b; split; trivial.
    right; apply B2.
  Qed.
    
  Lemma set_map_intro : forall s a b f (PF : forall x y, x < y -> f x < f y),
    a = f b -> set_in b s ->
    set_in a (set_map s PF).
  Proof.
    destruct s.
    unfold set_in; simpl.
    induction s as [| x s]; intros.
    inversion H0.
    subst a.
    destruct H0.
    subst x; left; reflexivity.
    destruct (andb_true_eq _ _ (sym_eq e)).
    right.
    apply IHs with (b := b); auto.
  Qed.

End SetMap.


(** ******************************************************************* *)
(** set eqvalence *)
(** ******************************************************************* *)
Section SetEqv.

  Definition set_eqv (s1 s2 : set) := 
    forall n, set_in n s1 <-> set_in n s2.

  Definition sd_eqv (s1 s2 : set_data) := 
    forall n, sd_in n s1 <-> sd_in n s2.

  Lemma sd_eqv_refl : forall s, sd_eqv s s.
  Proof.
    red; intros.
    tauto.
  Qed.

  Lemma sd_eqv_sym : forall s1 s2,
    sd_eqv s1 s2 -> sd_eqv s2 s1.
  Proof.
    intros s1 s2 He n.
    hypreplace He (He n).
    tauto.
  Qed.

  Lemma sd_eqv_trans : forall s1 s2 s3,
    sd_eqv s1 s2 -> sd_eqv s2 s3 -> sd_eqv s1 s3.
  Proof.
    intros s1 s2 s3 e1 e2 n.
    hypreplace e1 (e1 n).
    hypreplace e2 (e2 n).
    tauto.
  Qed.

  Lemma sd_eqv_elim : forall s1 s2,
    sd_eqv (sd_of s1) (sd_of s2) -> set_eqv s1 s2.
  Proof.
    tauto.
  Qed.

  Lemma sd_eqv_intro : forall s1 s2,
    set_eqv s1 s2 -> sd_eqv (sd_of s1) (sd_of s2).
  Proof.
    tauto.
  Qed.
   
  Variable s s1 s2 s3 : set.

  Lemma set_eqv_refl : set_eqv s s.
  Proof.
    red; intros.
    tauto.
  Qed.

  Lemma set_eqv_sym : 
    set_eqv s1 s2 -> set_eqv s2 s1.
  Proof.
    intros He n.
    hypreplace He (He n).
    tauto.
  Qed.

  Lemma set_eqv_trans : 
    set_eqv s1 s2 -> set_eqv s2 s3 -> set_eqv s1 s3.
  Proof.
    intros e1 e2 n.
    hypreplace e1 (e1 n).
    hypreplace e2 (e2 n).
    tauto.
  Qed.
  
End SetEqv.

(** Setoid Theory : set_eqv *)
Definition set_S : Setoid_Theory set set_eqv.
  split; [ exact set_eqv_refl | 
    exact set_eqv_sym | exact set_eqv_trans ].
Qed.

Add Setoid set set_eqv set_S.

Add Morphism set_in : set_in_ext.
  intros.
  hypreplace H (H n).
  tauto.
Qed.

Add Morphism set_add : set_add_ext.
  split; intros.
  destruct (set_add_elim H0).
  subst n0; apply set_add_intro1.
  apply set_add_intro2.
  rewrite <- H; exact H1.
  destruct (set_add_elim H0).
  subst n0; apply set_add_intro1.
  apply set_add_intro2.
  rewrite H; exact H1.
Qed.
  
Add Morphism set_remove : set_remove_ext.
  split; intros.
  destruct (set_remove_elim H0).
  apply set_remove_intro2; trivial.
  rewrite <- H; exact H2.
  destruct (set_remove_elim H0).
  apply set_remove_intro2; trivial.
  rewrite H; exact H2.
Qed.

Add Morphism set_union : set_union_ext.
  split; intros.
  destruct (set_union_elim H1).
  apply set_union_intro1.
  rewrite <- H; exact H2.
  apply set_union_intro2.
  rewrite <- H0; exact H2.
  destruct (set_union_elim H1).
  apply set_union_intro1.
  rewrite H; exact H2.
  apply set_union_intro2.
  rewrite H0; exact H2.
Qed.

Add Morphism set_diff : set_diff_ext.
  split; intros.
  destruct (set_diff_elim H1).
  apply set_diff_intro.
  rewrite <- H; exact H2.
  rewrite <- H0; exact H3.
  destruct (set_diff_elim H1).
  apply set_diff_intro.
  rewrite H; exact H2.
  rewrite H0; exact H3.
Qed.


(** ******************************************************************* *)
(** set extensionality *)
(** ******************************************************************* *)
Require Import Classical_Prop.

Section Extensionality.

  Definition sd_eq (s s' : set) := (sd_of s) = (sd_of s').

  Lemma sd_eq_intro : forall s s',
    s = s' -> sd_eq s s'.
  Proof.
    intros.
    destruct s; destruct s'. 
    injection H; intros; subst s0.
    rewrite H. 
    unfold sd_eq; reflexivity.
  Qed.
  
  Lemma sd_eq_elim : forall s s',
    sd_eq s s' -> s = s'.
  Proof.
    intros.
    destruct s; destruct s'.
    unfold sd_eq in H.
    simpl in H.
    subst s0.
    assert (e = e0).
    apply proof_irrelevance.
    subst e0.
    reflexivity.
  Qed.

  Lemma sd_eqv_nil_eq : forall s,
    sd_eqv nil s -> s = nil.
  Proof.
    destruct s; intros.
    reflexivity.
    destruct (H n).
    elimtype False; apply H1; left; reflexivity.
  Qed.
  
  Lemma set_eqv_empty_eq : forall s,
    set_eqv empty_set s -> s = empty_set.
  Proof.
    intros.
    hypreplace H (sd_eqv_intro H).
    hypreplace H (sd_eqv_nil_eq H).
    apply sd_eq_elim.
    unfold empty_set; unfold sd_eq; simpl.
    exact H.
  Qed.

  Theorem set_ext : forall s s',
    set_eqv s s' -> s = s'.
  Proof.
    intro s.
    destruct s.
    induction s; intros.
    assert (pfset (s:=nil) e = empty_set).
    apply sd_eq_elim.
    unfold empty_set; unfold sd_eq; reflexivity.
    rewrite H0 in H; rewrite H0.
    symmetry.
    apply (set_eqv_empty_eq H).
    destruct s'.
    apply sd_eq_elim.
    unfold sd_eq; simpl.
    assert (H1 := sd_eqv_intro H).
    simpl in H1.
    simpl in e.
    destruct (andb_true_eq _ _ (sym_eq e)).
    assert (sd_eqv s (sd_remove a s0)).
    split; intros.
    apply sd_remove_intro2.
    hypreplace H0 (blt_sd_elim (sym_eq H0)).
    hypreplace H0 (H0 _ H3).
    omega.
    destruct (H1 n).
    apply H4; right; exact H3.
    destruct (sd_remove_elim H3).
    destruct (H1 n).
    destruct (H7 H5); intuition.
    assert (e1 := sd_remove_canon a e0).
    assert (e2 := sym_eq H2).
    assert (set_eqv (pfset (s:=s) e2) (pfset (s:=(sd_remove a s0)) e1)).
    apply sd_eqv_elim.
    apply H3.
    hypreplace H4 (IHs _ _ H4).
    assert (E := sd_eq_intro H4).
    unfold sd_eq in E; simpl in E.
    rewrite E.
    rewrite <- sd_insert_lt.
    apply sd_insert_remove; trivial.
    destruct (H1 a).
    apply H5; left; reflexivity.
    hypreplace H0 (blt_sd_elim (sym_eq H0)).
    red; intros.
    destruct (H3 x); auto.
  Qed.

End Extensionality.


(** ******************************************************************* *)
(** others *)
(** ******************************************************************* *)
Lemma set_union_empty_right : forall s, set_union s empty_set = s.
Proof. 
  intros; apply set_ext; split; intros.
  destruct (set_union_elim H); [trivial | destruct H0].
  apply set_union_intro1; assumption.
Qed.

Lemma set_union_empty_left : forall s, set_union empty_set s = s.
Proof. 
  intros; apply set_ext; split; intros.
  destruct (set_union_elim H); [destruct H0 | trivial].
  apply set_union_intro2; assumption.
Qed.

Ltac elimsetin' x T :=
  match goal with
    | [SS: set_in x (set_add ?y ?A) |- _] =>
      let Eq := fresh "Eq" in 
	let SS' := fresh "SS" in
	  (destruct (set_add_elim SS) as [Eq | SS']; 
	    try rewrite Eq; clear SS; elimsetin' x T)
    | [SS: set_in x (set_union ?A ?B) |- _] =>
      let SS' := fresh "SS" in
	(destruct (set_union_elim SS) as [SS' | SS']; 
	  clear SS; elimsetin' x T)
    | [SS : set_in x (set_remove ?y ?A) |- _] =>
      let Ne := fresh "Ne" in 
	let SS' := fresh "SS" in
	  (destruct (set_remove_elim SS) as [Ne SS']; 
	    clear SS; elimsetin' x T)
    | [SS : set_in x (set_diff ?A ?B) |- _] =>
      let SS1 := fresh "SS" in 
	let SS2 := fresh "SS" in
	  (destruct (set_diff_elim SS) as [SS1 SS2]; 
	    clear SS; elimsetin' x T)
    | [SS: set_in x empty_set |- _ ] =>
      inversion SS
    | _ => (T x)
  end.

Ltac idfun x := idtac.

Ltac elimsetin x := elimsetin' x idfun.

Hint Resolve 
  set_union_intro1 set_union_intro2
  set_add_intro1 set_add_intro2 
  set_remove_intro1 set_remove_intro2 
  set_diff_intro: setin.

Ltac simplset' T :=
  match goal with 
    | [|-context [(set_remove ?y ?A)]] => 
      match goal with
	| [|- set_in ?x _ ] => 
	  let R := fresh "R" in 
	    (compare x y; intro R; 
	      [(subst x || rewrite R) | idtac]; elimsetin' x T)
      end

    | [|-context [(set_union ?x empty_set)]] => 
      rewrite set_union_empty_right; elimsetin' x T
    | [|-context [(set_union empty_set ?x)]] => 
      rewrite set_union_empty_left; elimsetin' x T

    | [Eqv : (set_union ?A ?B) = ?C |- set_in ?x ?C ] => 
      rewrite <- Eqv; elimsetin' x T
    | [Eqv : ?C = (set_union ?A ?B) |- set_in ?x ?C ] => 
      rewrite Eqv; elimsetin' x T
    | [Eqv : (set_add ?A ?B) = ?C |- set_in ?x ?C ] => 
      rewrite <- Eqv; elimsetin' x T
    | [Eqv : ?C = (set_add ?A ?B) |- set_in ?x ?C ] => 
      rewrite Eqv; elimsetin' x T
    | [|- set_in ?x _ ] => elimsetin' x T
    | _ => idtac
  end.

Ltac autosetin x := auto with setin.

Ltac simplset := simplset' autosetin.


Section OtherProperties.

  Lemma set_add_length_inc : forall s a, 
    ~ set_in a s -> length (sd_of (set_add a s)) = S (length (sd_of s)).
  Proof.
    destruct s.
    induction s; intros.
    reflexivity.
    simpl.
    simplcmp a0 a.
    reflexivity.
    elimtype False; apply H.
    rewrite (beq_nat_eq _ _ (sym_eq RE)).
    left; reflexivity.
    simpl in * .
    destruct (andb_true_eq _ _ (sym_eq e)) as [B1 B2].
    rewrite IHs with (e := sym_eq B2); auto.
    intro F; apply H.
    right; apply F.
  Qed.

  Lemma set_remove_length_dec : forall s a, 
    set_in a s -> S (length (sd_of (set_remove a s))) = (length (sd_of s)).
  Proof.
    destruct s.
    induction s; intros.
    elimtype False; apply H.
    simpl in * .
    destruct (andb_true_eq _ _ (sym_eq e)) as [B1 B2].
    hypreplace B1 (lt_sd_not_in (blt_sd_elim (sym_eq B1))).
    destruct H.
    subst; simplbeq.
    rewrite (sd_remove_none B1); reflexivity.
    simplif.
    subst.
    tauto.
    simpl.
    rewrite (IHs (sym_eq B2) _ H).
    reflexivity.
  Qed.

  Lemma sd_remove_length_le : forall s a,
    length (sd_remove a s) <= length s.
  Proof.
    induction s; simpl; intros.
    auto.
    simplif.
    auto.
    hypreplace IHs (IHs a0).
    simpl; omega.
  Qed.

  Lemma sd_remove_length_lt : forall s a,
    sd_in a s -> length (sd_remove a s) < length s.
  Proof.
    induction s; intros.
    simpl in H; tauto.
    simpl in * ; destruct H.
    subst.
    simplbeq.
    generalize (sd_remove_length_le s a).
    omega.
    simplif; try subst; simpl;
      generalize (IHs _ H); omega.
  Qed.

  Lemma sd_eqv_length_le : forall s sd, 
    sd_eqv (sd_of s) sd -> length (sd_of s) <= length sd.
  Proof.
    induction s. 
    induction s; intros.
    simpl; auto with arith.
    simpl in * .
    destruct (andb_true_eq _ _ (sym_eq e)) as [B1 B2].
    hypreplace IHs (IHs (sym_eq B2)).      
    assert (sd_eqv s (sd_remove a sd)).
    hypreplace B1 (lt_sd_not_in (blt_sd_elim (sym_eq B1))).
    split; intros.
    destruct (H n).
    apply sd_remove_intro2.
    intro F; subst; auto.
    apply H1; right; auto.
    destruct (H n).
    destruct (sd_remove_elim H0).
    destruct (H2 H4); auto; tauto.
    hypreplace IHs (IHs _ H0).
    assert (sd_in a sd).
    destruct (H a).
    apply H1; left; auto.
    generalize (sd_remove_length_lt H1).
    omega.
  Qed.
  
  Lemma sd_eqv_sd_no_dup : forall sd s a,
    sd_eqv (sd_of s) (a :: sd) -> length (sd_of s) = length (a :: sd) ->
    ~ sd_in a sd.
  Proof.
    induction sd.
    auto.
    red; intros.
    simpl in * .
    destruct H1.
    subst.
    assert (length (sd_of s) <= length (a :: sd)).
    apply sd_eqv_length_le.
    split; intros; destruct (H n).
    destruct (H2 H1).
    subst; left; auto.
    destruct H4.
    subst; left; auto.
    right; auto.
    apply H3.
    destruct H1.
    left; auto.
    right; right; auto.
    simpl in * ; omega.
    
    assert (length (sd_of s) <= length (a :: sd)).
    apply sd_eqv_length_le.
    split; intros; destruct (H n).
    destruct (H3 H2).
    subst; right; auto.
    destruct H5.
    subst; left; auto.
    right; auto.
    apply H4.
    destruct H2.
    subst; right; left; auto.
    right; right; auto.
    simpl in * ; omega.
  Qed.

End OtherProperties.

Definition set_subset s s' := 
  forall x, set_in x s -> set_in x s'.

Lemma set_diff_superset : forall s s',
  set_subset s s' -> set_diff s s' = empty_set.
Proof.
  intros.
  apply set_ext; split; intros; simplset.
  hypreplace H (H _ SS).
  tauto.
Qed.

Lemma set_subset_trans : forall s1 s2 s3,
  set_subset s1 s2 -> set_subset s2 s3 -> set_subset s1 s3.
Proof.
  red; intros.
  apply (H0 _ (H _ H1)).
Qed.

Ltac setsolver := apply set_ext; split; intros; simplset.

Lemma set_in_dec : forall x s, set_in x s \/ ~set_in x s.
Proof.
  destruct s; intros.
  destruct (sd_in_dec s x); [left | right]; assumption.
Qed.

Lemma empty_set_intro : forall s,
  (forall x, ~ set_in x s) -> s = empty_set.
Proof.
  intros.
  apply set_ext; split; intros.
  elimtype False; apply (H _ H0).
  destruct H0.
Qed.

Notation set_disj := (fun (ss1 ss2 : set) => 
  (forall x, set_in x ss1 -> ~ set_in x ss2)).

(*   Definition sd1 := (1::3::nil). *)
(*   Definition pf1 : bsd_canon sd1 = true. *)
(*     reflexivity. *)
(*   Defined. *)

(*   Definition sd2 := (1::4::nil). *)
(*   Definition pf2 : bsd_canon sd2 = true. *)
(*     reflexivity. *)
(*   Defined. *)

(*   Definition set1 : set := pfset pf1. *)
(*   Definition set2 : set := pfset pf2. *)
  
(*   Eval compute in  (sd_of (set_diff set1 set2)). *)
