(** range nat set *)

(** Chunxiao Lin @ 2006 *)

Require Export natset2.
Require Import basicutil.

Set Implicit Arguments.
Unset Strict Implicit.


(** ******************************************************************* *)
(** {k | 0 <= k < n } *)
(** ******************************************************************* *)
Section SetZeroN.

  Fixpoint set0n (n : nat) : set :=
    match n with
      | O => empty_set
      | S n' => set_add n' (set0n n')
    end.
  
  Lemma set0n_intro1 : forall a b,
    a < b ->
    set_in a (set0n b).
  Proof.
    induction b.
    intro; elimtype False; omega.
    simpl.
    intro LE.
    inversion_clear LE;
      intuition.
  Qed.
  
  Lemma set0n_elim1 : forall b a,
    set_in a (set0n b) ->
    a < b.
  Proof.
    induction b; intros.
    inversion H.
    simpl in H.
    elimsetin a; intuition.
  Qed.
  
  Lemma set0n_intro2 : forall a b,
    a >= b ->
    ~ set_in a (set0n b).
  Proof.
    red; intros.
    hypreplace H0 (set0n_elim1 H0).
    omega.
  Qed.
  
  Lemma set0n_elim2 : forall b a,
    ~ set_in a (set0n b) ->
    a >= b.
  Proof.
    intros.
    assert (~ a < b);
      [intro F; apply H | omega].
    apply (set0n_intro1 F).
  Qed.
  
End SetZeroN.

Hint Resolve 
  set0n_intro1 set0n_intro2 : setin.


(** ******************************************************************* *)
(** {k | a <= k < b } *)
(** ******************************************************************* *)
Section SetAB.

  Definition setab (a b:nat) : set :=
    set_diff (set0n b) (set0n a).
  
  Lemma setab_intro1 : forall a c b,
    b <= a < c ->
    set_in a (setab b c).
  Proof.
    unfold setab.
    intros a c b [LE LT].
    apply set_diff_intro; auto with setin.
  Qed.

  Lemma setab_elim1 : forall a c b,
    set_in a (setab b c) ->
    b <= a < c.
  Proof.
    unfold setab.
    intros a c b I.
    elimsetin a.
    constructor; auto with setin.
    hypreplace SS0 (set0n_elim2 SS0).
    auto with arith.
    apply set0n_elim1; auto.
  Qed.

  Lemma setab_intro2 : forall a c b,
    (a < b \/ a >= c) ->
    ~ set_in a (setab b c).
  Proof.
    red; intros.
    hypreplace H0 (setab_elim1 H0).
    omega.
  Qed.

  Lemma setab_elim2 : forall a c b,
    ~ set_in a (setab b c) ->
    (a < b \/ a >= c).
  Proof.
    intros.
    assert (~ b <= a < c).
    intro F; apply H. 
    apply (setab_intro1 F).
    omega.
  Qed.

End SetAB.

Hint Resolve 
  setab_intro1 setab_intro2 : setin.

Ltac elimrange x T :=
  match goal with
    | [SS: set_in x (set0n ?y) |- _] =>
      hypreplace SS (set0n_elim1 SS); elimrange x T
    | [SS: ~ set_in x (set0n ?y) |- _] =>
      hypreplace SS (set0n_elim2 SS); elimrange x T
    | [SS: set_in x (setab ?a ?b) |- _] =>
      let SS1 := fresh "SS" in let SS2 := fresh "SS" in
	(destruct (setab_elim1 SS) as [SS1 SS2]; 
	  clear SS; elimrange x T)
    | [SS: ~ set_in x (set0n ?y) |- _] =>
      let SS1 := fresh "SS" in
	(destruct (setab_elim2 SS) as [SS1 | SS1]; 
	  clear SS; elimrange x T)
    | _ => (T x)
  end.

Ltac elimrangeauto x := elimrange x autosetin. (* elimrange x autosetin. *)

Ltac simplrange := simplset' elimrangeauto.

Hint Extern 4 (set_in ?x (set_add ?y ?A)) => 
  replace y with x;[ apply set_add_intro1 | omega] : setin.

Hint Extern 4 (?a <= ?b < ?c) => omega : setin.

Hint Extern 4 (?a < ?b ) => omega : setin.

Hint Extern 4 (?a >= ?b ) => omega : setin.


Lemma setab_sb : forall a b,
  a <= b ->
  setab a (S b) = set_add b (setab a b).
Proof.
  intros.
  apply set_ext; intro x.
  compare a b; intro R2; [subst b | idtac].
  split; intros; simplrange.
  compare x b; intro R1; [subst b | idtac];
    split; intros; simplrange.
Qed.

Lemma setab_empty : forall a,
  setab a a = empty_set.
Proof.
  intros.
  apply set_ext.
  intro x; constructor; intro I;
    simplrange.
  elimtype False; omega. 
Qed.
  

(** ******************************************************************* *)
(** set of addresses *)
(** ******************************************************************* *)
Lemma addr_fun_ok : forall regMin x y,
  x < y -> (fun k => regMin + 4 * k) x < (fun k => regMin + 4 * k) y.
Proof.
  intros.
  omega.
Qed.

Definition addrs (min max: nat) :=
  set_map (set0n (div4 (max - min))) (addr_fun_ok min).

Lemma addrs_intro : forall a x y k,
  4 + a <= y -> a = x + 4 * k ->
  set_in a (addrs x y).
Proof.
  intros.
  subst.
  unfold addrs.
  eapply set_map_intro.
  reflexivity.
  apply set0n_intro1.
  assert (4 * k + 4 <= y - x).
  omega.
  clear H.
  generalize dependent k.
  generalize dependent (y - x); clear y x.
  intro n; pattern n.
  apply ind4; intros; try omega.
  rewrite div4plus.
  destruct k.
  omega.
  assert (k < div4 n0).
  apply H.
  omega.
  omega.
Qed.

Lemma addrs_elim : forall a x y,
  set_in a (addrs x y) ->
  4 + a <= y /\ exists k, a = x + 4 * k.
Proof.
  intros a x y I.
  hypreplace I (set_map_elim I).
  destruct I as [b [bEq I]].
  subst.
  constructor.
  hypreplace I (set0n_elim1 I).
  assert (4 * b + 4 <= y - x).
  2:omega.
  generalize dependent b.
  generalize dependent (y - x); clear x y.
  intro n; pattern n; apply ind4;
    try (unfold div4; simpl; intros b H; omega).
  intros a IH b.
  rewrite div4plus.
  intros H.
  destruct b.
  simpl; omega.
  assert (4 * b + 4 <= a).
  apply IH.
  omega.
  omega.
  exists b.
  omega.
Qed.

Lemma addrs_remove_min : forall m n,
  set_remove m (addrs m n) = (addrs (m + 4) n).
Proof.
  intros.
  apply set_ext.
  split; intro Ha.
  destruct (set_remove_elim Ha).
  decompose [and ex] (addrs_elim H0).
  assert (x <> 0).
  intro e.
  subst x.
  subst n0.
  apply H.
  auto.
  apply addrs_intro with (k := x - 1);
    omega.
  decompose [and ex] (addrs_elim Ha).
  apply set_remove_intro2.
  omega.
  apply addrs_intro with (k := x + 1);
    omega.
Qed.


